/*
 cs_byte_swapping.h
  MyCSeries include for MyCSeries byte reversal routines (such as endian conversion)

  Version 0.4

  Created: March 19, 2004
  Revised: August 3, 2006


    Version 1.0 has not yet been reached.  If you have a name conflict,
  please feel free to contact me via E-mail with a list of alternative
  names, so the name conflict can be resolved by the next release.
*/

/*
 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 *  This source code is offered for use in the public domain.  You may use,
 * modify or distribute it freely.
 *
 *  This code is distributed in the hope that it will be useful but WITHOUT
 * ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY DISCLAIMED.
 *  This includes but is not limited to warranties of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.
 */

#ifndef __BYTE_SWAPPING_H
# define __BYTE_SWAPPING_H

  /* just -1 and -2 were used for _2byte/_4byte in the original CSeries */

  enum
  {
    /* swapping entry types */
     _end,      /* termination */
     _skip,     /* skip w/o byte swapping -- anything within (0,127] */
    /* byte swap (negative value of size) */
     _2byte         = -2,
     _4byte         = -4,
     _8byte         = -8,
     _16byte        = -16,
    /* up to 120 */

    /* opcodes */
    /* for arrays - in big endian format */
     _entry32       = 0x80, /* followed by a UInt32 (for >127 skips and >64 swaps) */
     _array8        = 0x81, /* followed by an 8-bit count, array entries, and _end */
     _array16       = 0x82, /* followed by a 16-bit count, array entries, and _end */
     _array32       = 0x83  /* followed by a 32-bit count, array entries, and _end */
  };

  typedef UInt8         _bs_field; /* the original cseries used "word" (UInt16) */

 /* the ones that return an `int' return (0) on failure, (1) on swap, and (2) for
   when the current is the desired (no need to swap) */

 /* byte-swaps a structure using `list' as a guide */
  EXTERN_API_C(int) byte_swap_data(
    UInt8 *current, /* if NULL, then it will be swapped regardless */
    UInt8 desired, /* swapped if current!=desired, `current' is then updated */
    void *data,
    long size, /* size of the in-going data */
    long count, /* count of entries in `list' */
    const _bs_field *list);
 /* byte-swaps a table using `entry' as the guide */
  EXTERN_API_C(void) byte_swap_memory(
    UInt8 *current, /* if NULL, then it will be swapped regardless */
    UInt8 desired, /* swapped if current!=desired, `current' is then updated */
    void *data,
    UInt8 entry, /* single entry type */
    long count); /* count of the entry */
 /* byte-swaps an array using `list' as the guide */
  EXTERN_API_C(int) byte_swap_array(
    UInt8 *current, /* if NULL, then it will be swapped regardless */
    UInt8 desired, /* swapped if current!=desired, `current' is then updated */
    void *data,
    long size,
    long repeats, /* number of repeats for the array */
    const _bs_field *list); /* _end terminated list of entries */

 /* generic byte reversal routine - returns the next address */
  EXTERN_API_C(void*) reverse_bytes(
    long size,
    void *data);

# define byte_swap_to_native(b,d,z,c,l)     byte_swap_data(b,TARGET_RT_LITTLE_ENDIAN ? 0:1,d,z,c,l)

#endif
