/*
 cs_graphics.h
  MyCSeries include for graphics and graphics devices

  Version 0.4

  Created: November 25, 2004
  Revised: June 14, 2008


    Version 1.0 has not yet been reached.  If you have a name conflict,
  please feel free to contact me via E-mail with a list of alternative
  names so the name conflict can be resolved by the next release.
*/

/*
 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 *  This source code is offered for use in the public domain.  You may use,
 * modify or distribute it freely.
 *
 *  This code is distributed in the hope that it will be useful but WITHOUT
 * ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY DISCLAIMED.
 *  This includes but is not limited to warranties of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.
 */

#ifndef __CS_GRAPHICS_H
# define __CS_GRAPHICS_H

# include <cseries.h>
# include <cs_system.h>

# if PRAGMA_ONCE
#  pragma once
# endif

# ifdef __cplusplus
   extern "C" {
# endif

# if PRAGMA_IMPORT
#  pragma import on
# endif

# if PRAGMA_STRUCT_ALIGN
#  pragma options align=mac68k
# elif PRAGMA_STRUCT_PACKPUSH
#  pragma pack(push,2)
# elif PRAGMA_STRUCT_PACK
#  pragma pack(2)
# endif

# ifdef TARGET_OS_MAC
#  ifndef __QUICKDRAW__
#   include <QuickDraw.h>
#  endif
# endif

/* ------- constants */

/* default width/height (it isn't a constant) */
# define DEFAULT_SCROLLBAR_WIDTH        16

# define PIXEL8_MAXIMUM_COLORS          256

# define PIXEL15_MAX_COMPONENT          31
# define PIXEL16_MAX_COMPONENT          63
# define PIXEL24_MAX_COMPONENT          255
# define PIXEL30_MAX_COMPONENT          1023
# define PIXEL30_COMPONENT_WHITE        510

  enum
  {
    /* graphics errors */
     kGR_mem_for_bitmap_update  = -200, /* not enough memory to update a bitmap/pixmap */
     kGR_mem_for_render         = -201, /* not enough memory to render something */
     kGR_missing_menubar_code   = -202, /* code for specific menubar (or any) not found */
     kGR_region_overflow        = -203, /* region exceeded maximum length */
     kGR_bm_unsupported_for_op  = -204, /* system software doesn't support operation for bitmap/pixmap */
     kGR_no_temporary_memory    = -205, /* could not allocate temporary data */
     kGR_no_structure_memory    = -206, /* could not allocate structure data */

     kGR_color_search_failed    = -210, /* could not find any color within tolerance */
     kGR_ctab_zilch_error       = -211, /* no colors in the table! */
     kGR_entry_protected        = -212, /* the color table entry is protected - find something else to do */

     kGR_invalid_device         = -220, /* graphics device type is invalid */
     kGR_bad_itab_resolution    = -221, /* bad value passed for inverse table resolution */
     kGR_bad_pixel_format       = -222, /* bad pixel format or depth */

    /* graphics video errors */
     kGV_driver_unsupported     = -250, /* system doesn't support driver or vice-versa (stinks either way) */
     kGV_device_timing_error    = -251, /* video driver or device doesn't support timing change */
     kGV_buffer_size_error      = -252, /* bad size or allocation problem */
     kGV_mirroring_unsupported  = -253, /* driver doesn't support mirroring */
     kGV_unknown_driver_error   = -254, /* unknown/invalid device driver error */
     kGV_no_zoomed_video        = -255  /* zoomed video unsupported */
  };

  enum
  {
    /* rectangle adjustment types */
     _adjust_rectangle_center   = 0,

    /* cs_system_colors - these may vary depending on the system */
     _system_window_highlight   = 0, /* light gray-blue on MacOS */
     _system_gray_blue,
     _system_gray15,
     _system_gray33,
     _system_gray47,
     _system_gray53,
     _system_gray67,
     _system_gray73,
     _system_gray87,
     _system_gray93,
     _system_alert_frame_color,
     _system_alert_content_color,
     _system_alert_tl_color, /* top-left of frame */
     _system_alert_br_color, /* bottom-right of frame */

     NUMBER_OF_SYSTEM_COLORS,

    /* cs_basic_colors */
    /* for colors, `dark' is 53.3% and `light' is +20% */
     _basic_color_black         = 0,
     _basic_color_dark_gray, /* 20% */
     _basic_color_dark_blue,
     _basic_color_light_blue,
     _basic_color_dark_green,
     _basic_color_light_green,
     _basic_color_dark_cyan,
     _basic_color_light_cyan,
     _basic_color_dark_red,
     _basic_color_light_red,
     _basic_color_dark_magenta,
     _basic_color_light_magenta,
     _basic_color_dark_yellow,
     _basic_color_light_yellow,
     _basic_color_light_gray, /* 80% */
     _basic_color_white,
    /* 16 - end of _pixel_format_basic_4 colors */
     _basic_color_red,
     _basic_color_orange_red,
     _basic_color_red_orange,
     _basic_color_orange,
     _basic_color_orange_yellow,
     _basic_color_yellow_orange,
     _basic_color_yellow,
     _basic_color_green_yellow,
     _basic_color_yellow_green, /* 24 */
     _basic_color_green,
     _basic_color_green_cyan,
     _basic_color_cyan_green,
     _basic_color_cyan,
     _basic_color_cyan_blue,
     _basic_color_blue_cyan,
     _basic_color_blue,
     _basic_color_magenta_blue, /* 32 */
     _basic_color_blue_magenta,
     _basic_color_magenta,
     _basic_color_red_magenta,
     _basic_color_magenta_red,
     _basic_color_maroon,
     _basic_color_violet,
     _basic_color_sky_blue,
     _basic_color_gray_blue, /* 40 */
     _basic_color_gray_green,
     _basic_color_gray_cyan,
     _basic_color_gray_red,
     _basic_color_gray_magenta,
     _basic_color_gray_yellow,
     _basic_color_gray, /* 40% gray */
     _basic_color_moderate_gray, /* 50.2% gray */
     _basic_color_concrete, /* 66.7% gray */
     _basic_color_marble, /* 93.3% gray */
     _basic_color_warm_marble,
     _basic_color_cool_marble,
     _basic_color_fine,
     _basic_color_fern,
     _basic_color_fog,
     _basic_color_chlorine,
     _basic_color_moss, /* 56 */
     _basic_color_olive,
     _basic_color_carnation,
     _basic_color_lavender,
     _basic_color_sepia,
     _basic_color_raw_sienna,
     _basic_color_dirt,
     _basic_color_tan,
    /* 64 - end of basic colors */

    /* alternative names for the basic colors */
     _basic_color_lime = _basic_color_light_green,
     _basic_color_lemon = _basic_color_light_yellow,
     _basic_color_tangerine = _basic_color_yellow_orange
  };

  enum
  {
    /* pixel formats */
     _pixel_format_default      = 0, /* only valid for surface/buffer creation */
     _pixel_format_i1           = 1, /* indexed 1 bit */
     _pixel_format_i2           = 2, /* indexed 2 bit */
     _pixel_format_i4           = 4, /* indexed 4 bit */
     _pixel_format_i6           = 6, /* indexed 6 bit -- packed (6666) */
     _pixel_format_i8           = 8, /* indexed 8 bit */
     _pixel_format_i12          = 12, /* indexed 12 bit -- packed (CC) */
     _pixel_format_1_5_5_5      = 15, /* 1:5:5:5 RGB or 5:5:5:1 BGR (system dependent) */
     _pixel_format_5_6_5        = 16, /* 5:6:5 RGB/BGR */
     _pixel_format_8_8_8        = 24, /* 8:8:8 RGB/BGR */
     _pixel_format_10_10_10     = 30, /* 10:10:10 RGB (red is actually 12 bits) */
     _pixel_format_8_8_8_8      = 32, /* 8:8:8:8 ARGB/BGRA/RGBA/ABGR */

     _pixel_format_grayscale    = 128, /* only valid for surface/buffer creation */
     _pixel_format_g1           = 129, /* black and white */
     _pixel_format_g2           = 130, /* 4 grays */
     _pixel_format_g3           = 131, /* 8 grays -- packed (33333333) */
     _pixel_format_g4           = 132, /* 16 grays */
     _pixel_format_g6           = 134, /* 64 bit -- packed (6666) */
     _pixel_format_g8           = 136, /* 256 grays */
     _pixel_format_g10          = 136, /* 512 normal grays, 512 overflow grays */
     _pixel_format_g12          = 140, /* 4096 grays -- packed (CC) */
     _pixel_format_g16          = 144, /* 16 bit gray */
     _pixel_format_ag16_16      = 160, /* 16 bit alpha, 16 bit gray */
     _pixel_format_g32s         = 224, /* 32 bit floating-point single */

     _pixel_format_yuv          = 256, /* only valid for surface/buffer creation */
     _pixel_format_1_1_1        = 259, /* 1:1:1 YUV */
     _pixel_format_4_1_1        = 262, /* 4:1:1 Interleaved */
     _pixel_format_2_1_1        = 264, /* YUV 2:1:1 */
     _pixel_format_yvu_9        = 265, /* YVU9 Planar */
     _pixel_format_4_4_4        = 268, /* Y'CbCr 8-bit 4:4:4 */
     _pixel_format_s4_2_2       = 271, /* YUV 4:2:2 byte ordering 16-signed */
     _pixel_format_u4_2_2       = 272, /* YUV 4:2:2 byte ordering 16-unsigned */
     _pixel_format_4_2_0        = 326, /* 4:2:0 YUV */
     _pixel_format_4_4_4_4      = 228, /* Y'CbCrA 8-bit 4:4:4:4 */
     _pixel_format_yv_4_2_2     = 336, /* YVYU 4:2:2 */

     _pixel_format_cmy          = 384, /* only valid for surface/buffer creation */
     _pixel_format_cmy_24       = 408, /* 8 bit cyan, magenta, yellow */
     _pixel_format_cmyk_32      = 416, /* 8 bit cyan, magenta, yellow, black */

     _pixel_format_graphics     = 512, /* only valid for surface/buffer creation */
     _pixel_format_basic_4      = 516, /* uses the cs_basic_colors table (4 bit) */
     _pixel_format_basic_8      = 520, /* uses the cs_basic_colors table (8 bit) */
     _pixel_format_argb84       = 544, /* seperated ARGB (1*8)*4 */
     _pixel_format_acmy84       = 608, /* seperated ACMY (1*8)*4 */

     _pixel_format_lookup       = 4096, /* only valid for color/inverse tables */
     _pixel_format_t4_4_4       = 4100, /* [16] 4:4:4 RGB (system endian/order) */
     _pixel_format_t5_5_5       = 4101, /* [16] 5:5:5 RGB (system endian/order) */
     _pixel_format_t5_6_5       = 4102, /* [16] 5:6:5 RGB (system endian/order) */
     _pixel_format_t8_8_8       = 4104, /* [32] 8:8:8 RGB (system endian/order) */
     _pixel_format_t11_11_10    = 4107, /* [32] 11:11:10 RGB (system endian/order) */

    /* color_to_pixel methods [threshold.8] [method.8]
     the minimum tolerance (and the multiplier for the field) is 0.5% -- the
      practical maximum would be 199 (99.996%) - anything higher would inevitably
      cause the search routine to return anything: for _tolerant, the first color

     NOTE: they may return -1 or the color count if none fall below threshold
    */
     _translate_color_tolerant  = 0, /* fastest; during search, stops below threshold */
     _translate_color_moderate,      /* fastest complete search */
     _translate_color_closest,       /* slowest - accurate gray-threshold calculations */
     _translate_color_gray,          /* returns a gray if none are within the threshold */
     _translate_color_brightest,     /* sticks to higher-bound */
     _translate_color_darkest,       /* sticks to lower-bound */
     _translate_color_blochy         /* sticks to the intensity of each component */
  };

  enum
  { /* cs_inverse_table flags */
     _itab_gray_table           = 0x80, /* color table has no saturation */
     _itab_may_be_shared        = 0x40  /* set if low on memory - clear before disposal */
  };

/* ------- structures */

  typedef struct
  {
     UInt8              a, r, g, b;
  } ARGB32;

  typedef struct cs_short_point
  {
     SInt16             x, y;
  } cs_short_point;

  typedef struct cs_long_point
  {
     SInt32             x, y;
  } cs_long_point;

  typedef struct cs_rgb_color
  {
     UInt16             red, green, blue;
  } cs_rgb_color;

  typedef struct cs_color_table/*(1544)*/
  {
     SInt16             flags, color_count;
     SInt32             seed; /* unique identifier for the table */
     cs_rgb_color       colors[256];
  } cs_color_table, cs_ctable_ptr;

  typedef struct cs_inverse_table/*(8+)*/
  {
    /* reservedx are for future use - set to zero */
     UInt8              flags, resolution, reserved1, reserved2;
     SInt32             seed; /* should be either the seed in the table or NONE */

     UInt8              table[1];
  } cs_inverse_table, *cs_itable_ptr;

/* ------- macros and typedefs */

# define CS_RGB_COMPONENT_TO_15(x)      ((31*(x)+(UINT16_MAX/2))/UINT16_MAX)
# define CS_RGB_COMPONENT_TO_24(x)      ((255*(x)+(UINT16_MAX/2))/UINT16_MAX)
# define CS_RGB_COMPONENT_TO_30(x)      ((510*(x)+(UINT16_MAX/2))/UINT16_MAX)

 /* old names */
# define UINT16_TO_DEPTH15              CS_RGB_COMPONENT_TO_15
# define UINT16_TO_DEPTH24              CS_RGB_COMPONENT_TO_24
# define UINT16_TO_DEPTH30              CS_RGB_COMPONENT_TO_30

# define PIN_FORM_PIXEL15(r,g,b)        _PIXEL15(MIN((unsigned)(r),31),MIN((unsigned)(g),31),MIN((unsigned)(b),31))
# define PIN_FORM_PIXEL24(r,g,b)        _PIXEL24(MIN((unsigned)(r),255),MIN((unsigned)(g),255),MIN((unsigned)(b),255))

 /*
   in the 12-10-10 format, the two high bits are meant for overflow prevention
   - they are a part of the red channel -- this was meant primarily for averaging
   30 bit pixels as raw UInt32s without producing loss

   make sure that 30 bit pixels are written to disk Big-Endian for compatibility
 */
# define _RED30(x)                      MIN((x)>>20,1023)
# define _GREEN30(x)                    (1023&((x)>>10))
# define _BLUE30(x)                     (1023&(x))

# define RED30(x)                       (1023&((x)>>20))
# define GREEN30                        _GREEN30
# define BLUE30                         _BLUE30

# define _PIXEL30(r,g,b)                ((UInt32)((((r)<<20)&0x3FF00000)|(((g)<<10)&0xFFC00)|((b)&1023)))

/*
  commented out because pixel16 has no definitive format

#define _PIXEL16(r,g,b) NATIVE_TO_LEU16(((UInt16)((r)&0x1F)|(((g)<<5)&0x7E0)|(((b)<<11)&0xF800)))

#define _PIXEL16(r,g,b) (((UInt16)((r)&0x1F)|(((g)<<5)&0x7E0)|(((b)<<11)&0xF800)))
#define FORM_PIXEL16 _PIXEL16
*/

# if TARGET_OS_WIN32
#  define _PIXEL15(r,g,b)               ((UInt16)((((r)<<1)&62)|(((g)<<6)&0x7C0)|(((b)<<11)&0xF800)))
#  define _PIXEL24(r,g,b)               ((UInt32)((((r)<<8)&0xFF00)|(((g)<<16)&0xFF0000)|((b)<<24)))
#  define _PIXEL32(r,g,b,a)             ((UInt32)(((r)&255)|(((r)<<8)&0xFF00)|(((g)<<16)&0xFF0000)|((b)<<24)))

#  define FORM_PIXEL15                  _PIXEL15
#  define FORM_PIXEL24                  _PIXEL24

#  define ALPHA15(x)                    (1&(x))
#  define RED15(x)                      (31&((x)>>1))
#  define GREEN15(x)                    (31&((x)>>6))
#  define BLUE15(x)                     (31&((x)>>11))

#  define PIXEL24_RED_OFFSET            2
#  define PIXEL24_GREEN_OFFSET          1
#  define PIXEL24_BLUE_OFFSET           0

#  define RED24(x)                      (255&(x))
#  define GREEN24(x)                    (255&((x)>>8))
#  define BLUE24(x)                     (255&((x)>>16))

#  define PIXEL32_RED_OFFSET            3
#  define PIXEL32_GREEN_OFFSET          2
#  define PIXEL32_BLUE_OFFSET           1

#  define ALPHA32(x)                    (255&(x))
#  define RED32(x)                      (255&((x)>>8))
#  define GREEN32(x)                    (255&((x)>>16))
#  define BLUE32(x)                     (255&((x)>>24))
# else
#  define _PIXEL15(r,g,b)               ((UInt16)((((r)<<10)&0x7C00)|(((g)<<5)&0x3E0)|((b)&31)))
#  define _PIXEL24(r,g,b)               ((UInt32)((((r)<<16)&0xFF0000)|(((g)<<8)&0xFF00)|((b)&255)))
#  define _PIXEL32(r,g,b,a)             ((UInt32)(((a)<<24)|(((r)<<16)&0xFF0000)|(((g)<<8)&0xFF00)|((b)&255)))

#  ifdef env68k
#   pragma parameter __D1 ALHPA15(__D0)
     extern SInt32 ALHPA15( UInt16 ) = {0xE9C0, 0x1401};
#   pragma parameter __D1 RED15(__D0)
     extern SInt32 RED15( UInt16 ) = {0xE9C0, 0x1445};
#   pragma parameter __D1 GREEN15(__D0)
     extern SInt32 GREEN15( UInt16 ) = {0xE9C0, 0x1585};

#   pragma parameter __D1 RED24(__D0)
     extern SInt32 RED24( UInt32 ) = {0xE9C0, 0x1208};
#   pragma parameter __D1 GREEN24(__D0)
     extern SInt32 GREEN24( UInt32 ) = {0xE9C0, 0x1408};

#   pragma parameter __D1 ALPHA32(__D0)
     extern SInt32 ALPHA32( UInt32 ) = {0xE9C0, 0x1408};
#   pragma parameter __D1 RED32(__D0)
     extern SInt32 RED32( UInt32 ) = {0xE9C0, 0x1208};
#   pragma parameter __D1 GREEN32(__D0)
     extern SInt32 GREEN32( UInt32 ) = {0xE9C0, 0x1408};

#   pragma parameter __D0 FORM_PIXEL15(__D2, __D1, __D0)
    extern UInt16 FORM_PIXEL15( UInt8, UInt8, UInt8 )
      = {0xEFC0, 0x1585, 0xEFC0, 0x2445};
#   pragma parameter __D0 FORM_PIXEL24(__D2, __D1, __D0)
    extern UInt32 FORM_PIXEL24( UInt8, UInt8, UInt8 )
      = {0xEFC0, 0x1408, 0xEFC0, 0x2208};
#   pragma parameter __D0 FORM_PIXEL30(__D2, __D1, __D0)
    extern UInt32 FORM_PIXEL30( UInt16, UInt16, UInt16 )
      = {0xEFC0, 0x130A, 0xEFC0, 0x208A};

#   undef PIN_FORM_PIXEL15
#   undef PIN_FORM_PIXEL24

#   pragma parameter __D0 PIN_FORM_PIXEL15(__D2, __D1, __D0)
    extern UInt16 PIN_FORM_PIXEL15( UInt16, UInt16, UInt16 )
      = {0xC40, 31, 0x6202, 0x701F, 0xC41, 31, 0x6202, 0x721F, 0xC42, 31,
         0x6202, 0x741F, 0xEFC0, 0x1585, 0xEFC0, 0x2445};
#   pragma parameter __D0 PIN_FORM_PIXEL24(__D2, __D1, __D0)
    extern UInt32 PIN_FORM_PIXEL24( UInt16, UInt16, UInt16 )
      = {0xC40, 255, 0x6202, 0x70FF, 0xC41, 255, 0x6202, 0x72FF, 0xC42, 255,
         0x6202, 0x74FF, 0xEFC0, 0x1408, 0xEFC0, 0x2208};
#  else
#   define ALPHA15(x)                   (1&(x>>15))
#   define RED15(x)                     (31&(x>>10))
#   define GREEN15(x)                   (31&(x>>5))

#   define RED24(x)                     (255&(x>>16))
#   define GREEN24(x)                   (255&(x>>8))

#   define ALPHA32(x)                   (255&(x>>24))
#   define RED32(x)                     (255&(x>>16))
#   define GREEN32(x)                   (255&(x>>8))

#   define FORM_PIXEL15                 _PIXEL15
#   define FORM_PIXEL24                 _PIXEL24
#   define FORM_PIXEL30                 _PIXEL30
#  endif

/* the only difference on MacOS is the alpha channel is ignored in 32 bit
  depth under QuickDraw, but still we have to treat it as if it was 3 bytes */
#  define PIXEL24_RED_OFFSET            0
#  define PIXEL24_GREEN_OFFSET          1
#  define PIXEL24_BLUE_OFFSET           2
#  define PIXEL32_RED_OFFSET            1
#  define PIXEL32_GREEN_OFFSET          2
#  define PIXEL32_BLUE_OFFSET           3

#  define BLUE15(x)                     (31&(x))
#  define BLUE24(x)                     (255&(x))
#  define BLUE32(x)                     (255&(x))
# endif

/*
 _CS_RGB8 is for generating cs_rgb_color constants in raw from 24 bit pixel components,
 _CS_RGB4 is for 12 bit components, and _CS_RGB is for floating point component values
*/
# define _CS_RGB8(r,g,b)                {(r)*257,(g)*257,(b)*257}
# define _CS_RGB4(r,g,b)                {(r)*4369,(g)*4369,(b)*4369}
# define _CS_RGB(r,g,b)                 {(r)*65535,(g)*65535,(b)*65535}

# define SET_CSCT_COLOR_COUNT(t,c)      do{(t)->color_count=(c);}while(0)
# define GET_CSCT_COLOR_COUNT(t)        (0+(t)->color_count)

# ifdef __cplusplus
   inline void INTERPOLATE_ARGB32(ARGB32 *d, ARGB32 *s, UInt8 A, UInt8 R, UInt8 G, UInt8 B)
   {  short_int _alpha = A;
      (d)->r = (R)+((((short_int)(_alpha))*(short_int)((short_int)((s)->r)-(R)))>>8);
      (d)->g = (G)+((((short_int)(_alpha))*(short_int)((short_int)((s)->g)-(G)))>>8);
      (d)->b = (B)+((((short_int)(_alpha))*(short_int)((short_int)((s)->b)-(B)))>>8);
   }
# else
#  define INTERPOLATE_ARGB32(d,s,A,R,G,B)   do{ short_int _alpha = A; \
    (d)->r = (R)+((((short_int)(_alpha))*(short_int)((short_int)((s)->r)-(R)))>>8); \
    (d)->g = (G)+((((short_int)(_alpha))*(short_int)((short_int)((s)->g)-(G)))>>8); \
    (d)->b = (B)+((((short_int)(_alpha))*(short_int)((short_int)((s)->b)-(B)))>>8);}while(0)
# endif

# ifdef TARGET_OS_MAC
   typedef Rect                     platform_rect;

   typedef RGBColor                 platform_rgb;
   typedef ColorSpec                platform_color;
   typedef CTabHandle               platform_color_table, platform_ctab;

   typedef GDHandle                 platform_grafdevice;
   typedef CGrafPtr                 platform_grafport;
   typedef WindowRef                platform_grafwindow;

#  ifdef __cplusplus
    inline DEFINE_API(int)
      RECT_WIDTH(platform_rect *r) {return ((r)->right)-((r)->left);}
    inline DEFINE_API(int)
      RECT_HEIGHT(platform_rect *r) {return ((r)->bottom)-((r)->top);}
    inline DEFINE_API(int)
      RECT_LEFT(platform_rect *r) {return (r)->left;}
    inline DEFINE_API(int)
      RECT_TOP(platform_rect *r) {return (r)->top;}

    inline DEFINE_API(void)
      SET_RECTANGLE(platform_rect *r, SInt16 l, SInt16 t, UInt16 rw, UInt16 rh)
    {(r)->right=(rw)+((r)->left=(l)), (r)->bottom=(rh)+((r)->top=(t));}

    inline DEFINE_API(void)
      OFFSET_RECTANGLE(platform_rect *r, int dx, int dy)
    {(r)->left+=(dx), (r)->top+=(dy), (r)->right+=(dx), (r)->bottom+=(dy);}

    inline DEFINE_API(void)
      SET_COLOR24(platform_rgb &c, UInt8 cr, UInt8 cg, UInt8 cb)
    {(c).red=(cr)*257; (c).green=(cg)*257; (c).blue=(cb)*257;}

    inline DEFINE_API(void)
      SET_COLOR48(platform_rgb &c, UInt16 cr, UInt16 cg, UInt16 cb)
    {(c).red=(cr); (c).green=(cg); (c).blue=(cb);}
#  else
#   define RECT_WIDTH(r)                ((int)((r)->right)-((r)->left))
#   define RECT_HEIGHT(r)               ((int)((r)->bottom)-((r)->top))
#   define RECT_LEFT(r)                 ((int)(r)->left)
#   define RECT_TOP(r)                  ((int)(r)->top)

/*#define OFFSET_RECTANGLE(r,h,v) OffsetRect(r,h,v)*/
#   define SET_RECTANGLE(r,l,t,rw,rh)   do{(r)->right=(rw)+((r)->left=(l)), (r)->bottom=(rh)+((r)->top=(t));}while(0)
#   define OFFSET_RECTANGLE(r,dx,dy)    do{ SInt16 _tmpx=dx,_tmpy=dy; (r)->left+=_tmpx, (r)->top+=_tmpy, (r)->right+=_tmpx, (r)->bottom+=_tmpy;}while(0)

#   define SET_COLOR24(c,cr,cg,cb)      do{(c).red=(cr)*257; (c).green=(cg)*257; (c).blue=(cb)*257;}while(0)
#   define SET_COLOR48(c,cr,cg,cb)      do{(c).red=(cr); (c).green=(cg); (c).blue=(cb);}while(0)
#  endif
# else
   struct MacRect { SInt16 top, left, bottom, right; };

#  ifdef SDL
#   define RGBColor                 rgb_color;    /* (identical) */
    typedef SDL_Rect                platform_rect;
    typedef SDL_Color               platform_rgb, platform_color;
    typedef SDL_Palette             *platform_color_table, *platform_ctab;

#   ifdef __cplusplus
     inline DEFINE_API(int)
       RECT_WIDTH(platform_rect *r) {return ((r)->w);}
     inline DEFINE_API(int)
       RECT_HEIGHT(platform_rect *r) {return ((r)->h);}
    inline DEFINE_API(int)
      RECT_LEFT(platform_rect *r) {return (r)->x;}
    inline DEFINE_API(int)
      RECT_TOP(platform_rect *r) {return (r)->y;}

     inline DEFINE_API(void)
       SET_RECTANGLE(platform_rect *r, SInt16 l, SInt16 t, UInt16 rw, UInt16 rh)
     {(r)->x=(l), (r)->y=(t), (r)->w=(rw), (r)->h=(rh);}

     inline DEFINE_API(void)
       OFFSET_RECTANGLE(platform_rect *r, SInt16 dx, SInt16 dy)
     {(r)->x+=(dx), (r)->y+=(dy);}

     inline DEFINE_API(void)
       SET_COLOR24(SDL_Color &c, UInt8 cr, UInt8 cg, UInt8 cb)
     {(c).r=(cr); (c).g=(cg); (c).b=(cb);}

     inline DEFINE_API(void)
       SET_COLOR48(SDL_Color &c, UInt16 cr, UInt16 cg, UInt16 cb)
     {(c).r=((cr)+127)/257; (c).g=((cg)+127)/257; (c).b=((cb)+127)/257;}
#   else
#    define RECT_WIDTH(r)               ((int)((r)->w))
#    define RECT_HEIGHT(r)              ((int)((r)->h))
#    define RECT_LEFT(r)                ((int)(r)->x)
#    define RECT_TOP(r)                 ((int)(r)->y)

#    define SET_RECTANGLE(r,l,t,rw,rh)  do{(r)->x=(l), (r)->y=(t), (r)->w=(rw), (r)->h=(rh);}while(0)
#    define OFFSET_RECTANGLE(r,dx,dy)   do{(r)->x+=(dx), (r)->y+=(dy);}while(0)

#    define SET_COLOR24(c,cr,cg,cb)     do{(c).r=(cr); (c).g=(cg); (c).b=(cb);}while(0)
#    define SET_COLOR48(c,cr,cg,cb)     do{(c).r=((cr)+127)/257; (c).g=((cg)+127)/257; (c).b=((cb)+127)/257;}while(0)
#   endif

    typedef void*                   platform_grafdevice;
    typedef void*                   platform_grafport;
    typedef void*                   platform_grafwindow;
#  else
    typedef MacRect                 platform_rect;

    typedef cs_rgb_color            platform_rgb, platform_color;
    typedef cs_color_table          *platform_color_table, *platform_ctab;

    typedef void*                   platform_grafdevice;
    typedef void*                   platform_grafport;
    typedef void*                   platform_grafwindow;

#   ifdef __cplusplus
     inline DEFINE_API(int)
       RECT_WIDTH(platform_rect *r) {return ((r)->right)-((r)->left);}
     inline DEFINE_API(int)
       RECT_HEIGHT(platform_rect *r) {return ((r)->bottom)-((r)->top);}
    inline DEFINE_API(int)
      RECT_LEFT(platform_rect *r) {return (r)->left;}
    inline DEFINE_API(int)
      RECT_TOP(platform_rect *r) {return (r)->top;}

     inline DEFINE_API(void)
       SET_RECTANGLE(platform_rect *r, SInt16 l, SInt16 t, UInt16 rw, UInt16 rh)
     {(r)->right=(rw)+((r)->left=(l)), (r)->bottom=(rh)+((r)->top=(t));}

     inline DEFINE_API(void)
       OFFSET_RECTANGLE(platform_rect *r, SInt16 dx, SInt16 dy)
     {(r)->left+=(dx), (r)->top+=(dy), (r)->right+=(dx), (r)->bottom+=(dy);}

     inline DEFINE_API(void)
       SET_COLOR24(platform_rgb &c, UInt8 cr, UInt8 cg, UInt8 cb)
     {(c).red=(cr)*257; (c).green=(cg)*257; (c).blue=(cb)*257;}

     inline DEFINE_API(void)
       SET_COLOR48(platform_rgb &c, UInt16 cr, UInt16 cg, UInt16 cb)
     {(c).red=(cr); (c).green=(cg); (c).blue=(cb);}
#   else
#    define RECT_WIDTH(r)               ((int)((r)->right)-((r)->left))
#    define RECT_HEIGHT(r)              ((int)((r)->bottom)-((r)->top))
#    define RECT_LEFT(r)                ((int)(r)->left)
#    define RECT_TOP(r)                 ((int)(r)->top)

#    define SET_RECTANGLE(r,l,t,rw,rh)  do{(r)->right=(rw)+((r)->left=(l)), (r)->bottom=(rh)+((r)->top=(t));}while(0)
#    define OFFSET_RECTANGLE(r,dx,dy)   do{ SInt16 _tmpx=dx,_tmpy=dy; (r)->left+=_tmpx, (r)->top+=_tmpy, (r)->right+=_tmpx, (r)->bottom+=_tmpy;}while(0)

#    define SET_COLOR24(c,cr,cg,cb)     do{(c).red=(cr)*257; (c).green=(cg)*257; (c).blue=(cb)*257;}while(0)
#    define SET_COLOR48(c,cr,cg,cb)     do{(c).red=(cr); (c).green=(cg); (c).blue=(cb);}while(0)
#   endif /* defined(__cplusplus) */

#  endif /* !defined(SDL) -- else */
# endif /* !defined(TARGET_OS_MAC) -- else */

  typedef platform_handle           cs_itable_handle;

/* ------- globals */

  extern const platform_rgb     cs_rgb_black, cs_rgb_white;
  extern platform_rgb           cs_system_colors[];

  extern const cs_rgb_color     *cs_basic_colors;

 /* inverse table lookup tables - lists are terminated by NULL */
  extern UInt8                  *cs_basic_itabs[];

 /* for direct to window interface drawing (rather than opening more windows) */
  extern platform_grafport      cs_world_port;
  extern platform_grafdevice    cs_world_device;

/* ------- prototypes */

  EXTERN_API_C(void) adjust_rectangle(
    const platform_rect *outline,
    const platform_rect *inRect,
    platform_rect *outRect,
    int type);

  EXTERN_API_C(void) get_window_frame(
    platform_grafwindow window,
    platform_rect *screen_bounds);

 /* this calls get_current_graphics_mode for `mode':
     NONE uses the value returned by get_current_graphics_mode;
     zero forces a black&white cursor;
     if positive, it tries for the color cursor
 */
  EXTERN_API_C(int) set_mouse_cursor(
    int id, /* the accessable number that defines this cursor */
    int mode); /* can add theme in color depths */

  EXTERN_API_C(platform_ctab) build_platform_color_table(
    cs_color_table *mctab,
    UInt8 to_device); /* going to a graphics device? (pass non-zero for "true") */

 /* returns the platform specific error code */
  EXTERN_API_C(int) direct_set_colors(
    short first,
    short count,
    platform_grafdevice gdh,
    platform_color*);

 /* returns 1 on success, 0 if empty (still successful), -1 on failure */
  EXTERN_API_C(int) build_color_table(cs_color_table*, platform_ctab);

 /* returns the pixel value (or table index) given the color and the depth */
  EXTERN_API_C(SInt32) cs_rgb_color_to_pixel(
    cs_rgb_color color,
    int pixelFormat,
    int method,
    cs_color_table *ctab); /* can be NULL */

 /* returns the pixel value (or table index) given a ARGB32 and the depth */
  EXTERN_API_C(SInt32) ARGB32_to_pixel(
    ARGB32 color,
    int pixelFormat,
    int method,
    cs_color_table *ctab); /* can be NULL */

 /* converts and sorts the cs_color_table to a smaller table of type `tableFormat'
   this routine expects the out going table pointer to be able to hold at least
   count*4 bytes; `outIndices' is the order of the sorted list in respect to the
   passed color table
 */
  EXTERN_API_C(void) cs_build_ctable(
    int tableFormat, /* this is one of the last "pixel" formats */
    cs_color_table *ctab,
    void *outTable, /* either an array of UInt16 or UInt32 */
    UInt8 *outIndices); /* can be NULL */

 /* functions the user may need to implement */

 /* should return positive value if in color and zero if in monochrome;
   otherwise, an error code */
  EXTERN_API_C(int) get_current_graphics_mode(void);

# ifdef __cplusplus
   }
# endif

# if PRAGMA_STRUCT_ALIGN
#  pragma options align=reset
# elif PRAGMA_STRUCT_PACKPUSH
#  pragma pack(pop)
# elif PRAGMA_STRUCT_PACK
#  pragma pack()
# endif

# ifdef PRAGMA_IMPORT_OFF
#  pragma import off
# elif PRAGMA_IMPORT
#  pragma import reset
# endif
#endif /* !defined(__CS_GRAPHICS_H) */
