/*
 cs_strings.h

  Version 0.4

  Created: November 25, 2004
  Revised: November 25, 2006


    Version 1.0 has not yet been reached.  If you have a name conflict,
  please feel free to contact me via E-mail with a list of alternative
  names, so the name conflict can be resolved by the next release.
*/

/*
 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 *  This source code is offered for use in the public domain.  You may use,
 * modify or distribute it freely.
 *
 *  This code is distributed in the hope that it will be useful but WITHOUT
 * ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY DISCLAIMED.
 *  This includes but is not limited to warranties of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.
 */

#ifndef __CS_STRINGS_H
# define __CS_STRINGS_H

# include <cseries.h>
# include <string.h>

# if PRAGMA_ONCE
#  pragma once
# endif

# ifdef __cplusplus
   extern "C" {
# endif

# if PRAGMA_IMPORT
#  pragma import on
# endif

/* ------- prototypes */

/* memory "mem" extensions */

 /* like memchr, but returns the first character with any of the bits in "mask" */
  EXTERN_API_C(void*) memchrbit(const void*, int mask, size_t len);

 /* counts the characters within the block */
  EXTERN_API_C(long) memcnt(const void*, int chr, size_t len);

 /* returns the number of characters with any of the bits in "mask" */
  EXTERN_API_C(long) memcntbit(const void*, int mask, size_t len);

 /* remaps the UInt8 string using "table" */
  EXTERN_API_C(void) memrem(UInt8*, long count, const UInt8 *table);

/* table routines */

 /* table "tbl" comparison; does a translated string comparison,
   mapped string "b" comparison with (unaffected) string "a"
  NOTE: if you use the c-type array to do c-type comparison, diacritical
    characters may not be recognized
 */
  EXTERN_API_C(int) tblcmp(
    const char *a, /* can be NULL */
    const char *b, /* can be NULL */
    const char *table);

 /* like memchr, but returns the first character matching "e" given a table;
   particularly useful for ctype scanning
 */
  EXTERN_API_C(char*) tblchr(
    const char*,
    int e,
    size_t len,
    const char *table);

 /* like memchrbit, but returns the first character with any of the bits in
   "mask" given a table; particularly useful for ctype scanning
 */
  EXTERN_API_C(UInt8*) tblchrbit(
    const UInt8*,
    int mask,
    size_t len,
    const char *table);

 /* like memcnt, but returns the count of "e" using a given table;
   particularly useful for ctype scanning
 */
  EXTERN_API_C(long) tblcnt(
    const UInt8*,
    int e,
    size_t len,
    const char *table);

 /* like memcntbit, but returns the count of characters with any of the bits in
   "mask" given a table; particularly useful for ctype scanning
 */
  EXTERN_API_C(long) tblcntbit(
    const UInt8*,
    int mask,
    size_t len,
    const char *table);

 /* these copy the UInt8 string in "src" to "dst", using a remapping table */
  EXTERN_API_C(void) tblcpy(
    UInt8 *dst,
    const UInt8 *src,
    long count,
    const UInt8 *table);
 /* these copy the UIntxx string in "src" to "dst", using a remapping table;
   if an entry in the source string goes out of the range of "tableSize",
     "ovflValue" is copied instead
 */
  EXTERN_API_C(void) tblcpy2(
    UInt16 *dst,
    const UInt16 *src,
    long count, /* source length */
    const UInt16 *table,
    UInt16 tableSize, /* the number of entries in the table */
    UInt16 ovflValue); /* if an entry goes outside the table */
  EXTERN_API_C(void) tblcpy4(
    UInt32 *dst,
    const UInt32 *src,
    long count, /* source length */
    const UInt32 *table,
    UInt32 tableSize, /* the number of entries in the table */
    UInt32 ovflValue); /* if an entry goes outside the table */

/* string lists */

 /* as of November 25, 2006, all string list routines (on MacOS) interpret the count
   as a UInt16 */

 /* returns the count of strings in the platform string list with the particular ID */
  EXTERN_API_C(int)     countstrlist(short id);

 /* coppies the string in the platform string list with the particular ID */
  EXTERN_API_C(int)     getcstr(char *dst, short id, int x);
  EXTERN_API_C(UInt8*)  getpstr(void *dst, short id, int x);

/* string conversion and concatenation */

  EXTERN_API_C(UInt8*)  pstrcat(void*, const void*);
  EXTERN_API_C(UInt8*)  pstrcpy(void*, const void*);

 /* copies a C-string in "src" to "dst" converting it to a Pascal string */
  EXTERN_API_C(UInt8*) cpstrcpy(
    void *dst, /* can be NULL, can be (void*)NONE */
    const void *src, /* can be NULL, can be (void*)NONE */
    short maxLength); /* negative values default to 255 */

/* user printf routines */
/*  you can copy iocustom.c as a template (as long you state that you modified it) */

 /* C-string printf */
  EXTERN_API_C(char*) csprintf(char *buf, const char *fmt,...);
 /* Pascal-string printf */
  EXTERN_API_C(UInt8*) psprintf(void *buf, const char *fmt,...);

 /* data string printf -- also includes support for automatic endian conversion

   b = byte
   h = half; 16 bit for integers, 32 bit for floating-point
   Z = Big endian
   z = little endian

   returns the `data' and not `buf'
 */
  EXTERN_API_C(UInt8*) dsprintf(void *buf, const char *fmt, const void *data);

# ifdef __cplusplus
   }
# endif

# ifdef PRAGMA_IMPORT_OFF
#  pragma import off
# elif PRAGMA_IMPORT
#  pragma import reset
# endif

#endif /* defined(__CS_STRINGS_H) */
