/*
 cs_text.h
  MyCSeries include for text

  Version 0.4

  Created: November 25, 2004
  Revised: December 30, 2006


    Version 1.0 has not yet been reached.  If you have a name conflict,
  please feel free to contact me via E-mail with a list of alternative
  names, so the name conflict can be resolved by the next release.
*/

/*
 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 *  This source code is offered for use in the public domain.  You may use,
 * modify or distribute it freely.
 *
 *  This code is distributed in the hope that it will be useful but WITHOUT
 * ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY DISCLAIMED.
 *  This includes but is not limited to warranties of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.
 */

#ifndef __CS_TEXT_H
# define __CS_TEXT_H

# include <cseries.h>
# include <cs_graphics.h>
# include <cs_text_encodings.h>

# if PRAGMA_ONCE
#  pragma once
# endif

# ifdef __cplusplus
   extern "C" {
# endif

# if PRAGMA_IMPORT
#  pragma import on
# endif

/* ------- constants */

  enum
  {
    /* text encoding errors */
     kTS_encoding_not_found         = -300, /* nothing of this exists in the font */
     kTS_encoding_not_initialized   = -301,
     kTS_encoding_cannot_be_used    = -302, /* that encoding cannot be used in that context */

     kTS_font_not_found             = -310
  };

  enum
  {
    /* suggested cliptext margins */
     _cliptext_normal_margin        = 5     /* of course this varies with scale */
  };

/* ------- structures */

  enum
  {
   /* "style" bits */
     _ts_style_plain                = 0,
     _ts_style_bold                 = 0x01,
     _ts_style_italic               = 0x02,
     _ts_style_underlined           = 0x04,
     _ts_style_outlined             = 0x08, /* edge pixels only */
     _ts_style_strike               = 0x10, /* strike-through */
     _ts_style_overlined            = 0x20, /* line over rather than under (can use both) */
     _ts_style_fixed_width          = 0x80, /* constricts all characters to one width */

    /* "special" bits */
     _ts_special_show_invisibles    = 0x40, /* symbols like paragraphs and tabs are rendered */
     _ts_special_boxed              = 0x80, /* fills back-text rectangle at box opacity */

    /* "direction" values (for alignment) */
     _ts_direction_default          = 0x00,
     _ts_direction_from_center      = 0x01,
     _ts_direction_from_left        = 0xFE,
     _ts_direction_from_right       = 0xFF,

    /* only these bits represent the same bits in the QuickDraw stylefield */
     _ts_QuickDraw_face_bits        = 0x0F,

     _ts_special_user_bits          = 0x0F
  };

  typedef struct cs_fontspec/*(8)*/
  {
     SInt32         id; /* platform font ID - set by get_new_fontspec */
     UInt8          special, style;

    /* for the point size, here, zero is interpreted as "default"; negative values
       are reserved (but currently do nothing) */
     SInt16         size; /* this used to be unsigned */
  } cs_fontspec;

 /*
   WARNING: You MUST call initialize_textspec on every surface, screen, or port
    depth change before using this structure - or else!
 */
  typedef struct cs_textspec/*(72)*/
  {
     cs_fontspec    font;
     ARGB32         color;

    /*
      "hThickness" adds manual "bolding"; "softness" is a special effect
        for softening the text -- non-zero at the least renders it anti-aliased
    */
     UInt8          direction, softness, hThickness, vThickness;

    /* clipping bounds - must be within the pixel buffer's bounds or you'll be sorry! */
     platform_rect  bounds;

    /*
      how much extra space between characters, in fixed point size (not pixels);
        normally you would set this to FIXED_ONE -- zero would have the text
        compressed to the point that characters may literally touch each other
      this field would be used for justified alignment
    */
     sfixed         spaceExtra;

    /*
      these only apply if the "_ts_special_boxed" special bit is one;
      "boxColor" is a basic color because it is only meant for basic seperation
    */
     UInt8          boxColor, boxOpacity, boxExtra, boxFeather;

    /* these are for matrix and rotation effects;

      for normal rendering, set xslop to FRACT_ONE, and yslop to zero

      negative FRACT_ONE for xslop would flip the text horizontally
      -1 for yslop would flip the text vertically
    */
     sfract         xslop, yslop;

     SInt32         privates[8];
  } cs_textspec;

/* ------- globals */

extern const char   *cs_MacRoman_uppercase; /* uppercase remapping table */

/* ------- prototypes */

 /* graphics ports that use the cs_textspec type should automatically call this */
  EXTERN_API_C(int) initialize_textspec(
    cs_textspec*,
    void *pixelData,
    SInt32 pixelFormat, /* look at cs_graphics.h */
    cs_color_table *ctab, /* can be NULL (grayscale, zero is black) */
    int pitch); /* bytes per row */

 /* sets the current textspec in cases where NULL is passed to the routines */
  EXTERN_API_C(int) set_default_textspec(cs_textspec *spec);
 /* returns NULL if none is set */
  EXTERN_API_C(cs_textspec*) get_default_textspec(void);

 /* these use the current graphics port (can be a port supporting a cs_textspec) */
 /* with `_qd_bits'!=0, the special user bits are used to cover the QuickDraw face */
  EXTERN_API_C(void) get_port_font(cs_fontspec*, UInt8 save_qd_bits);
  EXTERN_API_C(void) set_port_font(const cs_fontspec*, UInt8 restore_qd_bits);

 /* returns the total width of the text using the current graphics port */
  EXTERN_API_C(long) get_text_width(
    const cs_fontspec*, /* NULL points to current graphics port settings */
    void *s, /* can be NULL */
    SInt32 len);

 /* string truncation - uses the current graphics port font settings */

  EXTERN_API_C(void*) cliptext(
    void *sptr,
    int *text_length,
    int method,
    sfixed maximum_width, sfixed margin);

# ifdef __cplusplus
   inline clip_cstr(char *s, int m, int w, int r)
   { int _len = strlen(s);  cliptext(s,&_len,w,r,m);     (s)[_len] = 0; }

   inline clip_pstr(char *s, int m, int w, int r)
   { int _len = s[0];       cliptext(&s[1],&_len,w,r,m); (s)[0] = _len; }
# else
#  define clip_cstr(s,m,w,r)                do{int _len=strlen(s); cliptext(s,&_len,w,r,m); (s)[_len]=0;}while(0)
#  define clip_pstr(s,m,w,r)                do{int _len=s[0]; cliptext(&s[1],&_len,w,r,m); (s)[0]=_len;}while(0)
# endif

 /* makes a spec from a `fnts' list entry; returns -1 and a default spec on failure */
 /* NOTE: as of version 0.3, resource's handle is unaffected (doesn't set for purging) */
  EXTERN_API_C(int) get_new_fontspec(cs_fontspec*, short fntsID, int x);

 /* CSeries textspec functions (custom and not platform specific) */
 /*  passing a NULL textspec parameter requires that the current graphics
    port supports and holds the current cs_textspecs (does nothing otherwise) */
 /*
   ts_textwidth
     returns the maximum horizontal span text width for every line
 */
  EXTERN_API_C(int) ts_textwidth(
    cs_textspec *spec,
    void *text, SInt32 encoding, int text_length);

 /*
   ts_drawtext
     draws the text in a horizontal span given "position"; updates position
 */
  EXTERN_API_C(int) ts_drawtext(
    cs_textspec *spec,
    cs_short_point *position,
    void *text, SInt32 encoding, int text_length);

 /*
   ts_hilitetext
     draws the text in a horizontal span given "position"; updates position
 */
  EXTERN_API_C(int) ts_hilitetext(
    cs_textspec *spec,
    cs_short_point *position,
    void *text, SInt32 encoding, int text_length);

# ifdef PRAGMA_IMPORT_OFF
#  pragma import off
# elif PRAGMA_IMPORT
#  pragma import reset
# endif

# ifdef __cplusplus
   }
# endif

#endif /* defined(__CS_TEXT_H) */
