/*
 cs_text_encodings.h
  text encodings recognized by MyCSeries

  Version 0.4

  Created: December 27, 2005

  (cs_text.h includes this file - read cs_text.h)
*/

/*
 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 *  This source code is offered for use in the public domain.  You may use,
 * modify or distribute it freely.
 *
 *  This code is distributed in the hope that it will be useful but WITHOUT
 * ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY DISCLAIMED.
 *  This includes but is not limited to warranties of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.
 */

/* ------- text encodings [function.8] [width.8] [script.8] [region.4] [system.4] */

#ifndef __CS_TEXT_ENCODINGS_H
# define __CS_TEXT_ENCODINGS_H

  enum
  {
     _text_function_normal          = 0,
     _text_function_none            = 255,
     _text_width_default            = 0,

     _text_system_default           = 0,
     _text_system_ISO_8859          = 1,
     _text_system_MacOS             = 2,
     _text_system_Windows           = 3,
     _text_system_DOS               = 4,
     _text_system_EBCDIC            = 5,
     _text_system_NextStep          = 6,
     _text_system_GB                = 7,
     _text_system_JIS               = 8,
     _text_system_Shift_JIS         = 9,
     _text_system_RFC               = 10,
     _text_system_KSC               = 11,
     _text_system_CNS               = 12,
     _text_system_ISO_UCS           = 13,
     _text_system_special           = 14,
     _text_system_ASCII             = 15
  };

  enum
  {
     /* NONE (-1) */
     _text_encoding_unknown         = -2,
     _text_encoding_invalid         = -3,
     _text_encoding_default         = -4, /* (for generation only) */

    /* (_text_function_xxxx<<24)|(_text_system_xxxx<<8)|(_text_encoding_xxxx) */

    /* Roman (Western) */
     _text_encoding_Roman           = 0,
     _text_encoding_ISO_8859_1      = 0x0001, /* ISO 8859-1 (Latin) */
     _text_encoding_MacRoman        = 0x0002,
     _text_encoding_Windows_1252    = 0x0003, /* Windows-1252 (Latin), Windows ANSI */
     _text_encoding_DOS_437         = 0x0004, /* DOS-437 U.S. (Latin) */
     _text_encoding_EBCDIC_US       = 0x0005,
     _text_encoding_US_ASCII        = 0x000F,
    /* Roman (European) */
     _text_encoding_European        = 0x0010,
     _text_encoding_ISO_8859_2      = 0x0011, /* ISO 8859-2 (Latin) */
     _text_encoding_MacCE           = 0x0012, /* Mac Central Europe */
     _text_encoding_Windows_1250    = 0x0013, /* Windows-1250 (Latin - Central Europe) */
     _text_encoding_DOS_850         = 0x0014, /* DOS-850 "multi-lingual" */
     _text_encoding_EBCDIC_CP037    = 0x0015, /* code page 037, extended EBCDIC (Latin-1 set) for US,Canada...*/
     _text_encoding_NextStepLatin   = 0x0016,

     _text_encoding_ISO_8859_3      = 0x0021, /* ISO 8859-3 */
     _text_encoding_ISO_8859_4      = 0x0031, /* ISO 8859-4 */

     _text_encoding_ISO_8859_9      = 0x0041, /* ISO 8859-9 */
     _text_encoding_ISO_8859_10     = 0x0051, /* ISO 8859-10 */
     _text_encoding_ISO_8859_11     = 0x0061, /* ISO 8859-11 */
     _text_encoding_ISO_8859_12     = 0x0071, /* ISO 8859-12 */
     _text_encoding_ISO_8859_13     = 0x0081, /* ISO 8859-13 */
     _text_encoding_ISO_8859_14     = 0x0091, /* ISO 8859-14 */
     _text_encoding_ISO_8859_15     = 0x00A1, /* ISO 8859-15 */
    /* Japanese */
     _text_encoding_Japanese        = 0x0100,
     _text_encoding_EUC_JP          = 0x0108,
     _text_encoding_ISO_2022_JP     = 0x0117,
     _text_encoding_ISO_2022_JP_3   = 0x0118,
     _text_encoding_ISO_2022_JP_2   = 0x0127,
     _text_encoding_ISO_2022_JP_1   = 0x0137,
    /* Traditional Chinese */
     _text_encoding_ChineseTrad     = 0x0200,
     _text_encoding_EUC_CN          = 0x0207,
     _text_encoding_Big5            = 0x020E,
     _text_encoding_HZ_GB_2312      = 0x020A, /* HZ (RFC 1842, for Chinese mail & news)*/
     _text_encoding_ISO2022_CN      = 0x0211,
     _text_encoding_ISO2022_CN_EXT  = 0x0221,
    /* Simplified Chinese */
     _text_encoding_GDB2312         = 0x190E,
    /* Korean */
     _text_encoding_Korean          = 0x0300,
     _text_encoding_EUC_KR          = 0x030B,
     _text_encoding_ISO2022_KR      = 0x0311,
    /* Thai */
     _text_encoding_Thai            = 0x1500,
     _text_encoding_EUC_TW          = 0x150C,
    /* Arabic */
     _text_encoding_Arabic          = 0x0400,
     _text_encoding_ISO_8859_6      = 0x0401, /* ISO 8859-6 */
    /* Hebrew */
     _text_encoding_Hebrew          = 0x0500,
     _text_encoding_ISO_8859_8      = 0x0501, /* ISO 8859-8 */
    /* Greek */
     _text_encoding_Greek           = 0x0600,
     _text_encoding_ISO_8859_7      = 0x0601, /* ISO 8859-7 */
     _text_encoding_Windows_1253    = 0x0603, /* Windows-1253 */
    /* Cyrillic */
     _text_encoding_ISO_8859_5      = 0x0701, /* ISO 8859-5 */
     _text_encoding_Windows_1251    = 0x0703, /* Windows-1251 */
     _text_encoding_K0I8_R          = 0x070E, /* Russian */

     _text_encoding_Devanagari      = 0x0900,
     _text_encoding_Gurmukhi        = 0x0A00,
     _text_encoding_Gujarati        = 0x0B00,
     _text_encoding_Oriya           = 0x0C00,
     _text_encoding_Bengali         = 0x0D00,
     _text_encoding_Tamil           = 0x0E00,
     _text_encoding_Telugu          = 0x0F00,
     _text_encoding_Kannada         = 0x1000,
     _text_encoding_Malayalam       = 0x1100,
     _text_encoding_Sinhalese       = 0x1200,
     _text_encoding_Burmese         = 0x1300,
     _text_encoding_Khmer           = 0x1400,
     _text_encoding_Laotian         = 0x1600,
     _text_encoding_Georgian        = 0x1700,
     _text_encoding_Armenian        = 0x1800,
     _text_encoding_Tibetan         = 0x1A00,
     _text_encoding_Mongolian       = 0x1B00,
     _text_encoding_Ethiopic        = 0x1C00,
     _text_encoding_Vietnamese      = 0x1E00,
     _text_encoding_ExtArabic       = 0x1F00, /* Extended Arabic */
     _text_encoding_Turkish         = 0x2300,
     _text_encoding_Farsi           = 0x8C00,
 /* Unicode */
     _text_encoding_MacUnicode      = 0x007E02,
     _text_encoding_UTF7            = 0x077E00,
     _text_encoding_UTF8            = 0x087E00,
     _text_encoding_Unicode32Bit    = 0x207E00,
     _text_encoding_UnicodeV1_1     = 0x08001D,
     _text_encoding_ISO10646_1993   = _text_encoding_UnicodeV1_1,
     _text_encoding_UnicodeV2       = 0x08030D, /* Korean Hangul*/
     _text_encoding_UnicodeV3_1     = 0x10001D,

     _text_encoding_uninterpreted   = 0x2000 /* symbolic only */
  };

#endif /*!defined(__CS_TEXT_ENCODINGS_H)*/
