/*
 cseries.h
  MyCSeries C source basic include

  Version 0.4.2

  Created: March 12, 2004
  Revised: November 24, 2008

  Written by Adam P.

 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 *  This source code is offered for use in the public domain.  You may use,
 * modify or distribute it freely.
 *
 *  This code is distributed in the hope that it will be useful but WITHOUT
 * ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY DISCLAIMED.
 *  This includes but is not limited to warranties of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.

Notes/Guidelines on compiling with MyCSeries
  int's should be 32 bit, and enum's should be sized as an `int'.

  If you are compiling with SDL, the SDL header includes should come _before_
    this file or any of the "cs_" headers.  The "SDL" parsing-type variable
    should only be defined by the SDL headers.
  If the MacTypes header is being used (this file may automatically include
    MacTypes when compiling for MacOS and/or QuickTime), make sure it is
    "included" before this file.

  The "cs_" files are _not_ automatically "included".  This file is meant
    _only_ for cross-platform compatibility, so it is not intended to contain
    prototypes for routines that may not work on all platforms.  It is more
    the other way around - the "cs_" files include THIS file.  This file is
    the bare minimum for MyCSeries.

  Some 68K inlines reside here to minimize library linkage.

WARNING:
  Some macros in this header (and the rest of the "cs_" headers) may use their
    parameters more than once.  Therefore, you should make sure that the
    parameters being passed are local variables that don't increment "(*a++)".
    Let the compiler worry about optimizations.

  For further details, see the README that comes in the MyCSeries package for
    this version (where this file came from).
*/

#ifndef __MY_CSERIES__
# define __MY_CSERIES__

# ifdef __CONDITIONALMACROS__
#  define COMPILER_HAS_LONGLONG     TYPE_LONGLONG
# else
#  if defined(__SUNPRO_C) || defined(__DECC)
  /* ConditionalMacros should not be included as it may not work */
#   if !defined(__STRICT_ANSI__)
#    define COMPILER_HAS_LONGLONG   1
#   else
#    define COMPILER_HAS_LONGLONG   0  /* not reliable under this condition */
#   endif

#   error Compiler or platform is not yet supported.
#  else
#   include <ConditionalMacros.h>
#   define COMPILER_HAS_LONGLONG    TYPE_LONGLONG
#  endif
# endif

# if PRAGMA_ONCE
#  pragma once
# endif

# ifdef __cplusplus
   extern "C" {
# endif

# if PRAGMA_IMPORT
#  pragma import on
# endif

# if PRAGMA_STRUCT_ALIGN
#  pragma options align=mac68k
# elif PRAGMA_STRUCT_PACKPUSH
#  pragma pack(push, 2)
# elif PRAGMA_STRUCT_PACK
#  pragma pack(2)
# endif

# include <SizeTDef.h>

# define MYCSERIES_VERSION          0x0042

# ifndef PREFER_FLOATING_POINT
#  ifdef powerc
#   define PREFER_FLOATING_POINT    1
#  else
#   define PREFER_FLOATING_POINT    0
#  endif
# endif

  /*
     pre-define the variable "__source__" to the designated source file name;
       "-d" for MPW (preferably with the -b2 option when compiling for 68K),
       `Options->Prefix' in Symantec C or CodeWarrior (I think)

     Example:
       const static char  __source__[] = __FILE__;
  */

  /*
    "env68k" is defined under Classic 68K (68020 or greater only)
    "envppc" is defined under PowerPC with CFM
    "envx86" is defined under 80x86
  */
# if !TARGET_CPU_PPC && TARGET_CPU_68K && !TARGET_RT_MAC_CFM
#  ifdef THINK_C
#   if __option(mc68020)
#    define env68k                  1
#   endif
#  else
#   define env68k                   1
#  endif

   typedef float                native_single_t;
   typedef long double          native_double_t;

/* macros and traps for env68k */

#  ifdef env68k

  /* this is not supplied in MacMemory.h! */

#   pragma parameter __D0 InlineGetPtrSize(__A0)
    EXTERN_API_C( long )
      InlineGetPtrSize(char *p) ONEWORDINLINE(0xA021);
#  endif

# elif TARGET_CPU_PPC && !TARGET_CPU_68K && TARGET_RT_MAC_CFM
#  define envppc                    1

   typedef float                native_single_t;
   typedef double               native_double_t;

# else
#  if TARGET_CPU_X86
#   define envx86                   1
#  endif

   typedef float                native_single_t;
   typedef double               native_double_t;
# endif

# if defined(__cplusplus)
#  define _ACCESSORS_ARE_INLINED    1
#  define _inline                   inline
# else
#  define _ACCESSORS_ARE_INLINED    0
#  define _inline                       
# endif

/* types */

# ifndef NORM_PTR_64
#  if !defined(__STRICT_ANSI__)
#   if defined(__osf__)
#    define NORM_PTR_64             1
#   endif
#  endif
# endif

# if !TARGET_OS_MAC && !defined(__MACTYPES__)
#  define UInt32                    Uint32
#  define UInt16                    Uint16
#  define UInt8                     Uint8
#  define SInt32                    Sint32
#  define SInt16                    Sint16
#  define SInt8                     Sint8
#  ifndef _SDL_types_h
#   ifndef H_MMBASIC
     typedef Uint32             unsigned int;
     typedef Uint16             unsigned short;
     typedef Uint8              unsigned char;
#   endif
  /* oh boy... */
    typedef Sint32              signed int;
    typedef Sint16              signed short;
    typedef Sint8               signed char;
#  endif
#  ifdef SDL_HAS_64BIT_TYPE
#   define UInt64                   Uint64
#   define SInt64                   Sint64
#  elif TARGET_RT_LITTLE_ENDIAN
#   if COMPILER_HAS_LONGLONG
     typedef UInt64             unsigned long long;
#   else
     typedef struct UInt64 { Uint32 lo;  Uint32 hi; } UInt64;
     typedef struct SInt64 { Uint32 lo;  Sint32 hi; } SInt64;
#   endif
#  else
#   if COMPILER_HAS_LONGLONG
     typedef SInt64             signed long long;
#   else
     typedef struct UInt64 { Uint32 hi;  Uint32 lo; } UInt64;
     typedef struct SInt64 { Sint32 hi;  Uint32 lo; } SInt64;
#   endif
#  endif
#  ifndef Float64
#   define Float64                  short double
#  endif
# else
#  include <MacTypes.h>
# endif

# if TARGET_RT_LITTLE_ENDIAN
#  ifndef TARGET_RT_BIG_ENDIAN
#   warn TARGET_RT_LITTLE_ENDIAN defined but not TARGET_RT_BIG_ENDIAN
#  endif

#  ifndef UINT128PREDEF
    typedef struct { UInt32 hi1;  UInt32 lo1, hi0, lo0; } UInt128;
#  endif
#  ifndef SINT128PREDEF
    typedef struct { SInt32 hi1;  UInt32 lo1, hi0, lo0; } SInt128;
#  endif
#  ifndef UINT256PREDEF
    typedef struct { UInt32 hi3;  UInt32 lo3, hi2, lo2, hi1, lo1, hi0, lo0; } UInt256;
#  endif
#  ifndef SINT256PREDEF
    typedef struct { SInt32 hi3;  UInt32 lo3, hi2, lo2, hi1, lo1, hi0, lo0; } SInt256;
#  endif
# else
#  if UINT128PREDEF
    typedef struct { UInt32 lo0, hi0, lo1;  UInt32 hi1; } UInt128;
#  endif
#  if SINT128PREDEF
    typedef struct { UInt32 lo0, hi0, lo1;  UInt32 hi1; } SInt128;
#  endif
#  if UINT256PREDEF
    typedef struct { UInt32 lo0, hi0, lo1, hi1, lo2, hi2, lo3;  UInt32 hi3; } UInt256;
#  endif
#  if SINT256PREDEF
    typedef struct { UInt32 lo0, hi0, lo1, hi1, lo2, hi2, lo3;  SInt32 hi3; } SInt256;
#  endif
# endif

  typedef SInt32                sfixed, sfract;
  typedef UInt32                pixel24, pixel32, ufixed, ufract;
  typedef UInt16                pixel15, pixel16;
  typedef SInt16                short_fixed;

/* `boolean' is kept this way for backward compatibility */
/*  (any of these can be overridden by #define) */
  typedef UInt8                 boolean, pixel8;

/* keep in mind that `byte' is no longer defined as 8 bits only here */
# ifndef BYTE_BIT_COUNT
#  define BYTE_BIT_COUNT            8
# endif

# if BYTE_BIT_COUNT==8
   typedef UInt8                byte;
# elif BYTE_BIT_COUNT==16
   typedef UInt16               byte;
# endif

# if TARGET_CPU_68K
#  define short_int                 short
# else
#  define short_int                 int
# endif

/* one of my types - 32 bit integer and fraction */
# ifndef _UFIXED32_
   typedef UInt64               UFixed32;
#  define _UFIXED32_
# endif
# ifndef _SFIXED32_
   typedef SInt64               SFixed32;
#  define _SFIXED32_
# endif

/* to avoid collision from <limits.h> */
# if !defined(__LIMITS__) && !defined(__LIMITS_H)
#  define __LIMITS__
#  define __LIMITS_H

#  define INT_MIN                   ((int)(-1L<<(8*sizeof(int)-1)))
#  define INT_MAX                   ((int)(~INT_MIN))

/* for unicode - CHAR_SIZE should be predefined */

#  if CHAR_SIZE==4
#   define CHAR_BIT                 32
#   define CHAR_MAX                 UINT32_MAX
#  elif CHAR_SIZE==2
#   define CHAR_BIT                 16
#   define CHAR_MAX                 UINT16_MAX
#  else
#   define CHAR_BIT                 8
#   define CHAR_MAX                 UINT8_MAX
#  endif

#  if COMPILER_HAS_LONGLONG
#   define kSINT64_MAX              SINT64_MAX
#   define kSINT64_MIN              SINT64_MIN
#   define SINT64_MAX               9223372036854775807LL
#   define SINT64_MIN               (-SINT64_MAX-1)
#   define UINT64_MAX               18446744073709551615ULL

#   define LLONG_MAX                SINT64_MAX
#   define LLONG_MIN                SINT64_MIN
#   define ULLONG_MAX               UINT64_MAX

#  else
#   if TARGET_RT_LITTLE_ENDIAN
#    define kSINT64_MAX             {0x7fffffff,0}
#    define kSINT64_MIN             {(long)0x80000000,-1L}
#   else
#    define kSINT64_MAX             {0,0x7fffffff}
#    define kSINT64_MIN             {-1L,(long)0x80000000}
#   endif
  /* so each can actually be recognized as a number */
#   define SINT64_MAX               9223372036854775807.0
#   define SINT64_MIN               -9223372036854775808.0
#   define UINT64_MAX               18446744073709551615.0
#  endif

#  define SINT32_MAX                2147483647
#  define SINT32_MIN                -2147483648
#  define SINT16_MAX                32767
#  define SINT16_MIN                -32768
#  define SINT8_MAX                 127
#  define SINT8_MIN                 -128

#  define UINT32_MAX                4294967295
#  define UINT16_MAX                65535
#  define UINT8_MAX                 255

#  ifdef NORM_WIDTH_64       /* a "long" is 64 bits and a "short" is 32 bits, etc. */
#   define ULONG_MAX                UINT64_MAX
#   define LONG_MIN                 SINT64_MIN
#   define LONG_MAX                 SINT64_MAX

#   define USHORT_MAX               UINT32_MAX
#   define SHORT_MIN                SINT32_MIN
#   define SHORT_MAX                SINT32_MAX
#  else

#   ifdef NORM_PTR_64        /* a "long" is the same size as a pointer (64 bits) */
#    define ULONG_MAX               UINT64_MAX
#    define LONG_MIN                SINT64_MIN
#    define LONG_MAX                SINT64_MAX
#   else
#    define ULONG_MAX               UINT32_MAX
#    define LONG_MIN                SINT32_MIN
#    define LONG_MAX                SINT32_MAX
#   endif

#   define USHORT_MAX               UINT16_MAX
#   define SHORT_MIN                SINT16_MIN
#   define SHORT_MAX                SINT16_MAX
#  endif

#  define SLONG_MIN                 LONG_MIN
#  define SLONG_MAX                 LONG_MAX

 /* old names */
#  define SHRT_MIN                  SHORT_MIN
#  define SHRT_MAX                  SHORT_MAX
#  define USHRT_MAX                 USHORT_MAX
# endif

# define FIXED_ONE                  ((sfixed)0x10000)
# define FIXED_ONE_HALF             ((sfixed)0x08000)
# define FIXED_FRACTIONAL_BITS      16
# define INTEGER_TO_FIXED(x)        ((x)<<FIXED_FRACTIONAL_BITS)
# define FIXED_INTEGERAL_PART(x)    ((x)>>FIXED_FRACTIONAL_BITS)

# define FRACT_ONE                  ((sfract)0x40000000)
# define FRACT_ONE_HALF             ((sfract)0x20000000)
# define FRACT_FRACTIONAL_BITS      30

/*
  Endian Conversion/Control

NOTE: The _NATIVE_TO_xExxx macros are meant for constants and are unsigned.
*/

# define _SWAP_ENDIAN_16(x)         ((((x)>>8)&255)|((x)<<8)&0xff00)
# define _SWAP_ENDIAN_32(x)         ((((x)<<24)&0xff000000)|(((x)<<8)&0xff0000)|(((x)>>8)&0xff00)|(((x)>>24)&255))

/* if you didn't know, these, of course, go both ways */
# define BEU16_TO_NATIVE            NATIVE_TO_BEU16
# define BES16_TO_NATIVE            NATIVE_TO_BES16
# define BEU32_TO_NATIVE            NATIVE_TO_BEU32
# define BES32_TO_NATIVE            NATIVE_TO_BES32
# define LEU16_TO_NATIVE            NATIVE_TO_LEU16
# define LES16_TO_NATIVE            NATIVE_TO_LES16
# define LEU32_TO_NATIVE            NATIVE_TO_LEU32
# define LES32_TO_NATIVE            NATIVE_TO_LES32

# if TARGET_RT_LITTLE_ENDIAN
#  define _ENDIANSWAP_LE16
#  define _NATIVE_TO_LE32
#  define _NATIVE_TO_BE16           _SWAP_ENDIAN_16
#  define _NATIVE_TO_BE32           _SWAP_ENDIAN_32
#  define NATIVE_TO_LEU16
#  define NATIVE_TO_LES16
#  define NATIVE_TO_LEU32
#  define NATIVE_TO_LES32
#  ifdef __cplusplus
    inline UInt16 NATIVE_TO_BEU16(UInt16 x) {(UInt16)_SWAP_ENDIAN_16(x);}
    inline SInt16 NATIVE_TO_BES16(SInt16 x) {(SInt16)_SWAP_ENDIAN_16(x);}
    inline UInt32 NATIVE_TO_BEU32(UInt32 x) {(UInt32)_SWAP_ENDIAN_32(x);}
    inline SInt32 NATIVE_TO_BES32(SInt32 x) {(SInt32)_SWAP_ENDIAN_32(x);}
#  else
#   define NATIVE_TO_BEU16(x)       ((UInt16)_SWAP_ENDIAN_16(x))
#   define NATIVE_TO_BES16(x)       ((SInt16)_SWAP_ENDIAN_16(x))
#   define NATIVE_TO_BEU32(x)       ((UInt32)_SWAP_ENDIAN_32(x))
#   define NATIVE_TO_BES32(x)       ((SInt32)_SWAP_ENDIAN_32(x))
#  endif
# else
#  define _NATIVE_TO_BE16
#  define _NATIVE_TO_BE32
#  define _NATIVE_TO_LE16           _SWAP_ENDIAN_16
#  define _NATIVE_TO_LE32           _SWAP_ENDIAN_32
#  define NATIVE_TO_BEU16
#  define NATIVE_TO_BES16
#  define NATIVE_TO_BEU32
#  define NATIVE_TO_BES32
#  ifdef env68k
#   pragma parameter __D0 _SWAP_U16(__D0)
    extern UInt16 _SWAP_U16(UInt16 x) = {0xE158};
#   pragma parameter __D0 _SWAP_S16(__D0)
    extern SInt16 _SWAP_S16(SInt16 x) = {0xE158};
/*
   ROL.W        #8,D0
*/
#   pragma parameter __D0 _SWAP_U32(__D0)
    extern UInt32 _SWAP_U32(UInt32 x) = {0xE158, 0x4840, 0xE158};
#   pragma parameter __D0 _SWAP_S32(__D0)
    extern SInt32 _SWAP_S32(SInt32 x) = {0xE158, 0x4840, 0xE158};
/*
   ROL.W        #8,D0
   SWAP         D0
   ROL.W        #8,D0
*/

#   define NATIVE_TO_LEU16          _SWAP_U16
#   define NATIVE_TO_LES16          _SWAP_S16
#   define NATIVE_TO_LEU32          _SWAP_U32
#   define NATIVE_TO_LES32          _SWAP_S32
#  else
#   ifdef __cplusplus
     inline UInt16 NATIVE_TO_LEU16(UInt16 x) {return (UInt16)_SWAP_ENDIAN_16(x);}
     inline SInt16 NATIVE_TO_LES16(SInt16 x) {return (SInt16)_SWAP_ENDIAN_16(x);}
     inline UInt32 NATIVE_TO_LEU32(UInt32 x) {return (UInt32)_SWAP_ENDIAN_32(x);}
     inline SInt32 NATIVE_TO_LES32(SInt32 x) {return (SInt32)_SWAP_ENDIAN_32(x);}
#   else
#    define NATIVE_TO_LEU16(x)      ((UInt16)_SWAP_ENDIAN_16(x))
#    define NATIVE_TO_LES16(x)      ((SInt16)_SWAP_ENDIAN_16(x))
#    define NATIVE_TO_LEU32(x)      ((UInt32)_SWAP_ENDIAN_32(x))
#    define NATIVE_TO_LES32(x)      ((SInt32)_SWAP_ENDIAN_32(x))
#   endif
#  endif
# endif

 /* "UNONE" is no longer defined; use "(UIntXXX)NONE" instead */
# define NONE                       (-1)

 /* generic macros */

# define KILO                       (1<<10)
# define MEG                        (1<<20)

# define ABS(x)                     ((x)<0?-(x):x)
# define SGN(x)                     ((x)<0?-1:((x)?1:0))
# define SSGN(x,y)                  ((signed)(x)<(signed)(y)?-1:((x)?1:0))
# define USGN(x,y)                  ((unsigned)(x)<(unsigned)(y)?-1:((x)?1:0))

# define PIN(x,y,z)                 ((x)<(y)?(y):((x)>(z)?(z):(x)))
# define CEILING(x,y)               ((x)<(y)?(x):(y))
# define FLOOR(x,y)                 ((x)>(y)?(x):(y))
# define MIN(x,y)                   ((x)<(y)?(x):(y))
# define MAX(x,y)                   ((x)>(y)?(x):(y))

# ifdef SWAP
#  ifndef __DEF_SWAP
#   define __DEF_SWAP
#  endif
# elif !defined(__DEF_SWAP)
#  define SWAP(x,y)                 do{if(sizeof(x)==1) {UInt8 _swaptmp=x; x=y; \
    y=_swaptmp;} else if(sizeof(x)==2) {SInt16 _swaptmp=x; x=y; y=_swaptmp;} \
    else if(sizeof(x)==4) {SInt32 _swaptmp=x; x=y; y=_swaptmp;} \
    else if(sizeof(x)==8) {UInt64 _swaptmp=*(UInt64*)&(x); x=y; \
    *(UInt64*)&(y)=_swaptmp;}}while(0)
# endif

# ifndef FSWAP
#  define FSWAP(x,y)                do{if(sizeof(x)==4) {float _swaptmp=x; x=y; \
    y=_swaptmp;} else if(sizeof(x)==8) {double _swaptmp=x; x=y; y=_swaptmp;} \
    else if(sizeof(x)>=10) {long double _swaptmp=x; x=y; y=_swaptmp;}}while(0)
# endif

# define FLAG(x)                    (1<<(x))
# define FLAG32(x)                  (1L<<(x))
# define TEST_FLAG16(x,y)           ((x)&(1<<(y)))
# define TEST_FLAG32(x,y)           ((x)&(1L<<(y)))

# define EXT_UINT(x,y)              ((x)&(1L<<(y))-1)
# define EXT_SINT(x,y)              (((SInt32)((long)(x)<<(32-(y))))>>(32-(y)))
# define iexts                      (((SInt32)((long)(x)<<(32-(__cs_itmpb=y))))>>(32-__cs_itmpb))

# define ipin(x,y,z)                ((__cs_itmpa=(x))<(__cs_itmpb=(y))?__cs_itmpb:(__cs_itmpa>(__cs_itmpc=(z)) ? __cs_itmpc:__cs_itmpa))
# define imin(x,y)                  ((__cs_itmpa=(x))<(__cs_itmpb=(y))?__cs_itmpa:__cs_itmpb)
# define imax(x,y)                  ((__cs_itmpa=(x))>(__cs_itmpb=(y))?__cs_itmpa:__cs_itmpb)

  EXTERN_API_C( UInt32 ) ClippedU64ToU32( UInt64 x );
  EXTERN_API_C( SInt32 ) ClippedS64ToS32( SInt64 x );
# if defined(__cplusplus)
    inline SInt32 ClippedS64ToS32(SInt64 x) {
       if (S64GtU32(x,SINT32_MAX)))
          return SINT32_MAX;
       else if (Int64Hi(x)<=(-1) && (Int64Hi(x)<(-1) || ((signed)Int64Lo(x))>=0))
          return SINT32_MIN;
       return Int64Lo(x);
    }
    inline UInt32 ClippedU64ToU32(UInt64 x) {
       return (-(Int64Hi(x)!=0))|Int64Lo(x);
    }
# endif

# if env68k
  /* 2006.11.25 - added TST instruction where needed due to the fact that we can't
    always expect the condition codes to reflect rD0 */

  /* abs natively co-exists in SInt32 definition in some compilers */
#  pragma parameter __D0 iabs(__D0)
   extern int iabs( int ) = {0x4A80, 0x6A02, 0x4480};
/* ; the eor method is four bytes longer and only matches the worst-case scenario
   ;  for this when it comes to performance
   TST.L        D0
   BPL.S        *+4
   NEG.L        D0
*/

#  pragma parameter __D0 isgn(__D0)
   extern SInt8 isgn( int ) = {0x4A80, 0x5DC0, 0x5EC1, 0x9001};
/*
   TST.L        D0
   SLT          D0
   SGT          D1
   SUB.B        D1,D0
*/

 /* comparative sign */

#  pragma parameter __D0 ssgn(__D0,__D1)
   extern SInt8 ssgn( int, int ) = {0xB081, 0x5DC0, 0x5EC1, 0x9001};
/*
   CMP.L        D1,D0
   SLT          D0
   SGT          D1
   SUB.B        D1,D0
*/
#  pragma parameter __D0 usgn(__D0,__D1)
   extern SInt8 usgn( unsigned int, unsigned int ) = {0xB081, 0x55C0, 0x52C1, 0x9001};
/*
   CMP.L        D1,D0
   SCS          D0
   SHI          D1
   SUB.B        D1,D0
*/

#  define SET_FLAG8(x,y,z)          _SET_BIT(&(x),7-(y),z)
#  define SET_FLAG16(x,y,z)         _SET_BIT(&(x),15-(y),z)
#  define SET_FLAG32(x,y,z)         _SET_BIT(&(x),31-(y),z)

#  pragma parameter _SET_BIT(__A0,__D0,__D1)
   extern void _SET_BIT( void*, char, UInt8 ) = {0xEFD0, 0x1801};
   /* BFINS D1,(A0){D0:1} */

#  undef iexts
#  pragma parameter __D0 iexts(__D0,__D1)
   extern long iexts( long, UInt8 ) = {0x7420, 0x9401, 0xEBC0, 0xCA1};
/*
  MOVEQ         #32,D2
  SUB.B         D1,D2
  BFEXTS        D0{D2:D1},D0
*/
#  define kZero64                   {0,0}
#  if TARGET_RT_LITTLE_ENDIAN
#   define kInt64U32(x)             {x,0}
#   define kInt64S32(x)             {x,((x)<0) ? -1:0}
#  else
#   define kInt64U32(x)             {0,x}
#   define kInt64S32(x)             {((x)<0) ? -1:0,x}
#  endif

#  define SetInt64(x,y,z)           do{(x).hi=(y),(x).lo=(z);}while(0)
#  define LSetInt64(x,y)            do{(x).hi=(y)>>32,(x).lo=(y);}while(0)
#  define XSetInt64(x,y)            do{(x).hi=(y)/4294967296.0;(x).lo=(y)-4294967296.0*(unsigned long)((y)/4294967296.0);}while(0)

#  if !defined(SAdd64) && !defined(UAdd64) && !defined(_SAdd64) && !defined(_UAdd64)
#   pragma parameter _SAdd64(__A0,__A1)
    extern void _SAdd64( SInt64*, const SInt64* ) = {0x9080, 0xD189, 0xD189};
#   pragma parameter _UAdd64(__A0,__A1)
    extern void _UAdd64( UInt64*, const UInt64* ) = {0x9080, 0xD189, 0xD189};
/*
   SUB.L        D0,D0
   ADDX.L       -(A1),-(A0)
   ADDX.L       -(A1),-(A0)
*/

#   define UAdd64(x,y)              _UAdd64(1+&(x),1+&(y))
#   define SAdd64(x,y)              _SAdd64(1+&(x),1+&(y))
#  endif

#  if !defined(SSub64) && !defined(USub64) && !defined(_SSub64) && !defined(_USub64)
#   pragma parameter _USub64(__A0,__A1)
    extern void _USub64( UInt64*, const UInt64* ) = {0x9080, 0x9189, 0x9189};
#   pragma parameter _SSub64(__A0,__A1)
    extern void _SSub64( SInt64*, const SInt64* ) = {0x9080, 0x9189, 0x9189};
/*
   SUB.L        D0,D0
   SUBX.L       -(A1),-(A0)
   SUBX.L       -(A1),-(A0)
*/

#   define USub64(x,y)              _USub64(1+&(x),1+&(y))
#   define SSub64(x,y)              _SSub64(1+&(x),1+&(y))
#  endif

#  if !defined(SNeg64) && !defined(UNeg64) && !defined(_SNeg64) && !defined(_UNeg64)
#   pragma parameter _UNeg64(__A0)
    extern void _UNeg64( UInt64* ) = {0x44A8, 4, 0x4090};
#   pragma parameter _SNeg64(__A0)
    extern void _SNeg64( SInt64* ) = {0x44A8, 4, 0x4090};
/*
   NEG.L        4(A0)
   NEGX.L       (A0)
*/

#   define UNeg64(x)                _UNeg64(&(x))
#   define SNeg64(x)                _SNeg64(&(x))
#  endif

/* add/sub Int64 (as two halves) to/from Int64 */
/* Takes D1:D2 as a whole */
#  pragma parameter _Add64Int(__A0,__D1,__D2)
   extern void _Add64Int( void*, UInt32, UInt32 ) = {0x2018, 0xD590, 0xD181, 0x2100};
/*
   MOVE.L       (A0)+,D0
   ADD.L        D2,(A0)
   ADDX.L       D1,D0
   MOVE.L       D0,-(A0)
*/
#  define Add64Int(x,y,z)           _Add64Int(&(x),y,z)

#  pragma parameter _Sub64Int(__A0,__D1,__D2)
   extern void _Sub64Int( void*, UInt32, UInt32 ) = {0x2018, 0x9590, 0x9181, 0x2100};
/*
   MOVE.L       (A0)+,D0
   SUB.L        D2,(A0)
   SUBX.L       D1,D0
   MOVE.L       D0,-(A0)
*/
#  define Sub64Int(x,y,z)           _Sub64Int(&(x),y,z)

/* add SInt32 to Int64 */
#  pragma parameter _Add64S32(__A0,__D2)
   extern void _Add64S32( void*, SInt32 ) =
     {0xEBC2, 0x1001, 0x2018, 0xD590, 0xD181, 0x2100};
/*
   BFEXTS       D2{0:1},D1
   MOVE.L       (A0)+,D0
   ADD.L        D2,(A0)
   ADDX.L       D1,D0
   MOVE.L       D0,-(A0)
*/
#  define Add64S32(x,y)             _Add64S32(&(x),y)

/* subtract SInt32 from Int64 */
#  pragma parameter _Sub64S32(__A0,__D2)
   extern void _Sub64S32( void*, SInt32 ) =
     {0xEBC2, 0x1001, 0x2018, 0x9590, 0x9181, 0x2100};
/*
   BFEXTS       D2{0:1},D1
   MOVE.L       (A0)+,D0
   SUB.L        D2,(A0)
   SUBX.L       D1,D0
   MOVE.L       D0,-(A0)
*/
#  define Sub64S32(x,y)             _Sub64S32(&(x),y)

#  define Add64U32(x,y)             Add64Int(x,0,y)
#  define Sub64U32(x,y)             Sub64Int(x,0,y)

#  ifdef THINK_C
#   pragma parameter _LShiftU64(__A0,__D2)
    extern void _LShiftU64( UInt64*, UInt8 ) =
      {0x242, 127, 0x2018, 0x2200, 0xC02, 31, 0x6314, 0x4290, 0x642, -64, 0x9301,
       0x642, 32, 0x8401, 0xE5A8, 0x2100, 0x6016, 0xE5A8, 0x2210, 0x2100, 0x2001,
       0xE5A9, 0x4442, 0x642, 32, 0xE4A8, 0x8198, 0x2081};
#   pragma parameter _LShiftS64(__A0,__D2)
    extern void _LShiftS64( SInt64*, UInt8 ) =
      {0x242, 127, 0x2018, 0x2200, 0xC02, 31, 0x6314, 0x4290, 0x642, -64, 0x9301,
       0x642, 32, 0x8401, 0xE5A8, 0x2100, 0x6016, 0xE5A8, 0x2210, 0x2100, 0x2001,
       0xE5A9, 0x4442, 0x642, 32, 0xE4A8, 0x8198, 0x2081};
/*
   ANDI.W       #127,D2
   MOVE.L       (A0)+,D0
   CMPI.B       #31,D2
   BLS.S        *+22    ; @1
   CLR.L        (A0)
   ADDI.W       #-64,D2
   SUBX.B       D1,D1
   ADDI.W       #32,D2
   OR.B         D1,D2
   LSL.L        D2,D0
   MOVE.L       D0,-(A0)
   BRA.S        *+24    ; end
@1 LSL.L        D2,D0
   MOVE.L       (A0),D1
   MOVE.L       D0,-(A0)
   MOVE.L       D1,D0
   LSL.L        D2,D1
   NEG.W        D2
   ADDI.W       #32,D2
   LSR.L        D2,D0
   OR.L         D0,(A0)+
   MOVE.L       D1,(A0)
*/

#  pragma parameter _RShiftU64(__A0,__D2)
   extern void _RShiftU64( UInt64*, UInt8 ) =
     {0x242, 127, 0x2010, 0x2200, 0xC02, 31, 0x6312, 0x4298, 0x642, -64, 0x9100,
      0x642, 32, 0x8400, 0xE4A9, 0x6012, 0xE4A8, 0x20C0, 0x2010, 0xE4A8, 0x4442,
      0x642, 32, 0xE5A9, 0x8280, 0x2081};
/*
   ANDI.W       #127,D2
   MOVE.L       (A0),D0
   MOVE.L       D0,D1
   CMPI.B       #31,D2
   BLS.S        *+20    ; @1
   CLR.L        (A0)+
   ADDI.W       #-64,D2
   SUBX.B       D0,D0
   ADDI.W       #32,D2
   OR.B         D0,D2
   LSR.L        D2,D1
   BRA.S        *+20    ; @2
@1 LSR.L        D2,D0
   MOVE.L       D0,(A0)+
   MOVE.L       (A0),D0
   LSR.L        D2,D0
   NEG.W        D2
   ADDI.W       #32,D2
   LSL.L        D2,D1
   OR.L         D0,D1
@2 MOVE.L       D1,(A0)
*/

#  pragma parameter _RShiftS64(__A0,__D2)
   extern void _RShiftS64( SInt64*, UInt8 ) =
     {0x242, 127, 0x2010, 0x2200, 0xC02, 31, 0x6316, 0x402, 32, 0xD080, 0x9180,
      0x20C0, 0xC02, 31, 0x6218, 0xE4A1, 0x2081, 0x6014, 0xE4A0, 0x20C0, 0x2010,
      0xE4A8, 0x4442, 0x642, 32, 0xE5A9, 0x8081, 0x2080};
/*
   ANDI.W       #127,D2
   MOVE.L       (A0),D0
   MOVE.L       D0,D1
   CMPI.B       #31,D2
   BLS.S        *+24    ; @1
   SUBI.B       #32,D2
   ADD.L        D0,D0
   SUBX.L       D0,D0
   MOVE.L       D0,(A0)+
   CMPI.B       #31,D2
   BHI.S        *+26    ; @2
   ASR.L        D2,D1
   MOVE.L       D1,(A0)
   BRA.S        *+22    ; end
@1 ASR.L        D2,D0
   MOVE.L       D0,(A0)+
   MOVE.L       (A0),D0
   LSR.L        D2,D0
   NEG.W        D2
   ADDI.W       #32,D2
   LSL.L        D2,D1
   OR.L         D1,D0
@2 MOVE.L       D0,(A0)
*/
#  else
    extern void _LShiftU64( UInt64*, UInt8 );
    extern void _LShiftS64( SInt64*, UInt8 );

    extern void _RShiftU64( UInt64*, UInt8 );
    extern void _RShiftS64( SInt64*, UInt8 );
#  endif
#  define LShiftU64(x,y)            _LShiftU64(&(x),y)
#  define LShiftS64(x,y)            _LShiftS64(&(x),y)

#  define RShiftU64(x,y)            _RShiftU64(&(x),y)
#  define RShiftS64(x,y)            _RShiftS64(&(x),y)

#  pragma parameter _DoubleU64(__A0)
   extern void _DoubleU64( UInt64* ) = {0x9080, 0xD188, 0xD188};
#  pragma parameter _DoubleS64(__A0)
   extern void _DoubleS64( SInt64* ) = {0x9080, 0xD188, 0xD188};
/*
   SUB.L        D0,D0
   ADDX.L       -(A0),-(A0)
*/
#  define DoubleU64(x)              _DoubleU64(1+&(x))
#  define DoubleS64(x)              _DoubleS64(1+&(x))

#  pragma parameter _HalfU64(__A0)
   extern void _HalfU64( UInt64* ) =
     {0x2018, 0x2210, 0xE288, 0xE291, 0x2081, 0x2100};
/*
   MOVE.L       (A0)+,D0
   MOVE.L       (A0),D1
   LSR.L        #1,D0
   ROXR.L       #1,D1
   MOVE.L       D1,(A0)
   MOVE.L       D0,-(A0)
*/
#  define HalfU64(x)                _HalfU64(&(x))

 /* now does it right */
#  pragma parameter _HalfS64(__A0)
   extern void _HalfS64( SInt64* ) =
     {0x7400, 0x2228, 4, 0x2018, 0x6A04, 0x5281, 0xD182, 0xE280, 0xE291, 0x2081,
      0x2100};
/*
   MOVEQ        #0,D2
   MOVE.L       4(A0),D1
   MOVE.L       (A0)+,D0
   BPL.S        *+6     ; @1
   ADDQ.L       #1,D1
   ADDX.L       D2,D0
@1 ASR.L        #1,D0
   ROXR.L       #1,D1
   MOVE.L       D1,(A0)
   MOVE.L       D0,-(A0)
*/
#  define HalfS64(x)                _HalfS64(&(x))

/* multiply a xInt64 by a UInt32 */
#  pragma parameter _UMul64U32(__A0,__D2)
   extern void _UMul64U32( UInt64*, UInt32 ) =
    {0x2018, 0x4C02, 0, 0x4C10, 0x2401, 0xD280, 0x2082, 0x2101};
#  pragma parameter _SMul64U32(__A0,__D2)
   extern void _SMul64U32( SInt64*, UInt32 ) =
    {0x2018, 0x4C02, 0, 0x4C10, 0x2401, 0xD280, 0x2082, 0x2101};
/*
   MOVE.L       (A0)+,D0
   MULU.L       D2,D0
   MULU.L       (A0),D1:D2
   ADD.L        D0,D1
   MOVE.L       D2,(A0)
   MOVE.L       D1,-(A0)
*/
#  define UMul64U32(x,y)            _UMul64U32(&(x),y)
#  define SMul64U32(x,y)            _SMul64U32(&(x),y)

/* multiply a xInt64 by a SInt32 */
#  pragma parameter _UMul64S32(__A0,__D2)
   extern void _UMul64S32( UInt64*, SInt32 ) =
     {0x2018, 0x4A82, 0x6A06, 0x4482, 0x4490, 0x4080, 0x4C02, 0, 0x4C10, 0x2401,
      0xD280, 0x2082, 0x2101};
#  pragma parameter _SMul64S32(__A0,__D2)
   extern void _SMul64S32( SInt64*, SInt32 ) =
     {0x2018, 0x4A82, 0x6A06, 0x4482, 0x4490, 0x4080, 0x4C02, 0, 0x4C10, 0x2401,
      0xD280, 0x2082, 0x2101};
/*
   MOVE.L       (A0)+,D0
   TST.L        D2
   BPL.S        *+8     ; @1
   NEG.L        D2
   NEG.L        (A0)
   NEGX.L       D0
@1 MULU.L       D2,D0
   MULU.L       (A0),D1:D2
   ADD.L        D0,D1
   MOVE.L       D2,(A0)
   MOVE.L       D1,-(A0)
*/
#  define UMul64S32(x,y)            _UMul64S32(&(x),y)
#  define SMul64S32(x,y)            _SMul64S32(&(x),y)

/* divide a UInt64 by a UInt32 - returns remainder */
#  pragma parameter __D2 _UDiv64U32R(__A0,__D2)
   extern UInt32 _UDiv64U32R( UInt64*, UInt32 ) =
     {0x2018, 0x2210, 0x2082, 0x4C50, 2, 0x4C50, 0x1402, 0x2081, 0x2100};
/*
   MOVE.L       (A0)+,D0
   MOVE.L       (A0),D1
   MOVE.L       D2,(A0)
   TDIVU.L      (A0),D2:D0
   DIVU.L       (A0),D2:D1
   MOVE.L       D1,(A0)
   MOVE.L       D0,-(A0)
*/
#  define UDiv64U32(x,y)            do{_UDiv64U32R(&(x),y);}while(0)
#  define UDiv64U32R(x,y,z)         do{z=_UDiv64U32R(&(x),y);}while(0)

/* divide a SInt64 by a SInt32 - returns remainder */
#  pragma parameter __D2 _SDiv64S32R(__A0,__D2)
   extern UInt32 _SDiv64S32R( SInt64*, SInt32 ) =
     {0x2228, 4, 0x2010, 0x6A04, 0x4481, 0x4080, 0xB598, 0x2082, 0x6A02, 0x4490,
      0x4C50, 2, 0x4C50, 0x1402, 0x4AA0, 0x6A06, 0x4482, 0x4481, 0x4080, 0x20C0,
      0x2081};
/*
   MOVEA.L      4(A0),D1
   MOVE.L       (A0),D0
   BPL.S        *+6     ; @1
   NEG.L        D1
   NEGX.L       D0
@1 EOR.L        D2,(A0)+
   MOVE.L       D2,(A0)
   BPL.S        *+4     ; @2
   NEG.L        (A0)
@2 TDIVU.L      (A0),D2:D0
   DIVU.L       (A0),D2:D1
   TST.L        -(A0)
   BPL.S        *+8     ; @3
   NEG.L        D2
   NEG.L        D1
   NEGX.L       D0
@3 MOVE.L       D0,(A0)+
   MOVE.L       D1,(A0)
*/
#  define SDiv64S32(x,y)            do{_SDiv64S32R(&(x),y);}while(0)
#  define SDiv64S32R(x,y,z)         do{z=_SDiv64S32R(&(x),y);}while(0)

/* unsigned greater than test */
#  pragma parameter __D0 _U64Gt(__A0,__A1)
   extern boolean _U64Gt( const UInt64*, const UInt64* ) =
     {0x2018, 0x2210, 0x2419, 0x9291, 0x56C1, 0x9182, 0x52C0, 0x6502, 0x8001};
/*
   MOVE.L       (A0)+,D0
   MOVE.L       (A0),D1
   MOVE.L       (A1)+,D2
   SUB.L        (A1),D1
   SNE          D1
   SUBX.L       D2,D0
   SHI          D0
   BCS.S        *+4
   OR.B         D1,D0
*/
#  define U64Gt(x,y)                _U64Gt(&(x),&(y))

/* unsigned greater than test */
#  pragma parameter __D0 _U64Ge(__A0,__A1)
   extern boolean _U64Ge( const UInt64*, const UInt64* ) =
    {0x2018, 0x2210, 0x2419, 0x9291, 0x9182, 0x54C0};
/*
   MOVE.L       (A0)+,D0
   MOVE.L       (A0),D1
   MOVE.L       (A1)+,D2
   SUB.L        (A1),D1
   SUBX.L       D2,D0
   SCC          D0
*/
#  define U64Ge(x,y)                _U64Ge(&(x),&(y))

/* signed greater than test */
#  pragma parameter __D0 _S64Gt(__A0,__A1)
   extern boolean _S64Gt( const SInt64*, const SInt64* ) =
    {0x2018, 0x2210, 0x2419, 0x9291, 0x56C1, 0x9182, 0x5EC0, 0x6D02, 0x8001};
/*
   MOVE.L       (A0)+,D0
   MOVE.L       (A0),D1
   MOVE.L       (A1)+,D2
   SUB.L        (A1),D1
   SNE          D1
   SUBX.L       D2,D0
   SGT          D0
   BLT.S        *+4
   OR.B         D1,D0
*/
#  define S64Gt(x,y)                _S64Gt(&(x),&(y))

/* signed greater than test */
#  pragma parameter __D0 _S64Ge(__A0,__A1)
   extern boolean _S64Ge( const SInt64*, const SInt64* ) =
    {0x2018, 0x2210, 0x2419, 0x9291, 0x9182, 0x5CC0};
/*
   MOVE.L       (A0)+,D0
   MOVE.L       (A0),D1
   MOVE.L       (A1)+,D2
   SUB.L        (A1),D1
   SUBX.L       D2,D0
   SGE          D0
*/
#  define S64Ge(x,y)                _S64Ge(&(x),&(y))

/* multiply UInt32 by UInt32 to yield UInt64 */
#  pragma parameter _UMul32(__A0,__D0,__D2)
   extern void _UMul32( UInt64*, UInt32, UInt32 ) = {0x4C00, 0x2401, 0x20C1, 0x2082};
/*
   MULU.L       D0,D1:D2
   MOVE.L       D1,(A0)+
   MOVE.L       D2,(A0)
*/
#  define UMul32(x,y,z)             _UMul32(&(x),y,z)

/* multiply SInt32 by SInt32 to yield SInt64 */
#  pragma parameter _SMul32(__A0,__D0,__D2)
   extern void _SMul32( SInt64*, SInt32, SInt32 ) = {0x4C00, 0x2C01, 0x20C1, 0x2082};
/*
   MULS.L       D0,D1:D2
   MOVE.L       D1,(A0)+
   MOVE.L       D2,(A0)
*/
#  define SMul32(x,y,z)             _SMul32(&(x),y,z)

/* divide the value of a UInt64 by a UInt32 to yield a UInt32 */
#  pragma parameter __D1 _UDiv32(__A0,__D2)
   extern UInt32 _UDiv32( UInt64*, UInt32 ) =
   {0x2018, 0x2210, 0xB480, 0x620A, 0x2041, 0x4C42, 1, 0x2001, 0x2208, 0x4C42,
    0x1400};
#  pragma parameter __D0 _UMod32(__A0,__D2)
   extern UInt32 _UMod32( UInt64*, UInt32 ) =
   {0x2018, 0x2210, 0xB480, 0x620A, 0x2041, 0x4C42, 1, 0x2001, 0x2208, 0x4C42,
    0x1400};
/*
   MOVE.L       (A0)+,D0
   MOVE.L       (A0),D1
   CMP.L        D0,D2
   BHI.S        *+12    ; @1
   MOVE.L       D1,A0
   DIVUL.L      D2,D1:D0
   MOVE.L       D1,D0
   MOVE.L       A0,D1
@1 DIVU.L       D2,D0:D1
*/
#  define UDiv32(x,y)               _UDiv32(&(x),y)
#  define UMod32(x,y)               _UMod32(&(x),y)

/* divide the value of a SInt64 by a SInt32 to yield a SInt32 (may overflow) */
#  pragma parameter __D1 _SDiv32(__A0,__D2)
   extern SInt32 _SDiv32( SInt64*, SInt32 ) =
   {0x2018, 0x2210, 0x2F00, 0x6A04, 0x4481, 0x4080, 0xB597, 0x2F02, 0x6A02, 0x4497,
    0x4C57, 2, 0x4C5F, 0x1402, 0x4A9F, 0x6A02, 0x4481};
/*
   MOVE.L       (A0)+,D0
   MOVE.L       (A0),D1
   MOVE.L       D0,-(SP)
   BPL.S        *+6     ; @1
   NEG.L        D1
   NEGX.L       D0
@1 EOR.L        D2,(SP)
   MOVE.L       D2,-(SP)
   BPL.S        *+4     ; @2
   NEG.L        (SP)
@2 TDIVU.L      (SP),D2:D0
   DIVU.L       (SP)+,D2:D1
   TST.L        (SP)+
   BPL.S        *+4     ; end
   NEG.L        D1
*/
#  pragma parameter __D2 _SMod32(__A0,__D2)
   extern SInt32 _SMod32( SInt64*, SInt32 ) =
   {0x2018, 0x2210, 0x2F00, 0x6A04, 0x4481, 0x4080, 0xB597, 0x2F02, 0x6A02, 0x4497,
    0x4C57, 2, 0x4C5F, 0x1402, 0x4A9F, 0x6A02, 0x4482};
/*
   MOVE.L       (A0)+,D0
   MOVE.L       (A0),D1
   MOVE.L       D0,-(SP)
   BPL.S        *+6     ; @1
   NEG.L        D1
   NEGX.L       D0
@1 EOR.L        D2,(SP)
   MOVE.L       D2,-(SP)
   BPL.S        *+4     ; @2
   NEG.L        (SP)
@2 TDIVU.L      (SP),D2:D0
   DIVU.L       (SP)+,D2:D1
   TST.L        (SP)+
   BPL.S        *+4     ; end
   NEG.L        D2
*/
#  define SDiv32(x,y)               _SDiv32(&(x),y)
#  define SMod32(x,y)               _SMod32(&(x),y)

#  pragma parameter __D0 _ClippedU64ToU32(__A0)
   extern UInt32 _ClippedU64ToU32( UInt64* ) =
   {0x2018, 0x4480, 0x9180, 0x8090};
/*
   MOVE.L       (A0)+,D0
   NEG.L        D0
   SUBX.L       D0,D0
   OR.L         (A0),D0
*/
#  define ClippedU64ToU32(x)        _ClippedU64ToU32(&(x))

/* bitwise */
#  define Int64Not(x)               do{(x).hi=!(x).hi,(x).lo=!(x).lo;}while(0)
#  define Int64BNot(x)              do{(x).hi=~(x).hi,(x).lo=~(x).lo;}while(0)
#  define Int64BOr(x,y,z)           do{(x).hi|=(UInt32)(y),(x).lo|=(UInt32)(z);}while(0)
#  define Int64BAnd(x,y,z)          do{(x).hi&=(UInt32)(y),(x).lo&=(UInt32)(z);}while(0)
#  define Int64BExOr(x,y,z)         do{(x).hi^=(UInt32)(y),(x).lo^=(UInt32)(z);}while(0)
#  define BitOr64(x,y)              Int64BOr(x,(y).hi,(y).lo)
#  define BitAnd64(x,y)             Int64BAnd(x,(y).hi,(y).lo)
#  define BitExOr64(x,y)            Int64BExOr(x,(y).hi,(y).lo)

/* equal */
#  define Int64Eq(x,y)              ((Int64Hi(x)==Int64Hi(y)) && (Int64Lo(x)==Int64Lo(y)))
#  define Int64EqU32(x,y)           (!(x).hi && (x).lo==(y))
/* greater than */
#  define U64GtU32(x,y)             ((x).lo>(y) || (x).hi)
#  define S64GtU32(x,y)             ((signed)((x).hi)>=0 && ((x).lo>(y) || (x).hi))
/* greater than or equal */
#  define U64GeU32(x,y)             ((x).lo>=(y) || (x).hi)
#  define S64GeU32(x,y)             ((signed)((x).hi)>=0 && ((x).lo>=(y) || (x).hi))
/* less than */
#  define U64LtU32(x,y)             ((x).lo<(y) && !(x).hi)
#  define S64LtU32(x,y)             ((signed)((x).hi)<0 || ((x).lo<(y) && !(x).hi))
/* less than or equal */
#  define U64LeU32(x,y)             ((x).lo<=(y) && !(x).hi)
#  define S64LeU32(x,y)             ((signed)((x).hi)<0 || ((x).lo<=(y) && !(x).hi))

/* zero, negative, positive */
#  define Int64Z(x)                 (!(x).hi && !(x).lo)
#  define Int64N(x)                 ((signed)((x).hi)<0)
#  define Int64P(x)                 (((signed)((x).hi)>=0 && (x).lo) || (signed)((x).hi)>0)

/* high 32 bits, low 32 bits, translate to long int, translate to long double */
#  define Int64Hi(x)                (0+(x).hi)
#  define Int64Lo(x)                (0+(x).lo)
#  define Int64ToL(x)               (long int)(0+(x).lo)
#  define Int64ToX(x)               (((long double)((x).hi)*4294967296.0)+((long double)(x).lo))

/* leading zero count (also known as integer log-2) */
#  pragma parameter __D1 LZCount32(__D0)
   extern int LZCount32( UInt32 ) = {0xEDC0, 0x1000};
   /* BFFFO D0{0:0},D1 */

/* fractional/`fixed' math */

#  pragma parameter __D1 NFract32Mul(__D0, __D2)
   extern SInt32 NFract32Mul( SInt32, SInt32 ) = {0x4C02, 0xC01};
   /* MULS.L D2,D1:D0 */

#  pragma parameter __D1 UFract32Mul(__D0, __D2)
   extern SInt32 UFract32Mul( UInt32, UInt32 ) =
     {0x4C02, 0x401, 0xD080, 0x7000, 0xD380};
/*
   MULU.L       D2,D1:D0
   ADD.L        D0,D0
   MOVEQ        #0,D0
   ADDX.L       D0,D1
*/

#  pragma parameter __D0 NFix30Mul(__D0, __D2)
   extern sfract NFix30Mul( sfract, sfract ) =
     {0x4C00, 0x2C00, 0xD482, 0xD180, 0xD482, 0xD180};
/*
   MULS.L       D0,D0:D2
   ADD.L        D2,D2
   ADDX.L       D0,D0
   ADD.L        D2,D2
   ADDX.L       D0,D0
*/

/* WARNING: Passing a denominator of zero to this division inlines _will_ generate
    a divide by zero exception.  Also, overflow will give bad results (result may
    not match the non-env68k versions). */
#  pragma parameter __D0 NFix30Div(__D0, __D2)
   extern sfract NFix30Div( sfract, sfract ) =
     {0xEBC0, 0x101E, 0xE498, 0x280, 0xC000, 0, 0x4C42, 0x0C01};
/*
   BFEXTS       D0{0:30},D1
   ROR.L        #2,D0
   ANDI.L       #0xC0000000,D0
   DIVS.L       D2,D1:D0
*/

#  pragma parameter __D0 UFix16Mul(__D0, __D2)
   extern ufixed UFix16Mul( ufixed, ufixed ) =
     {0x4C02, 0x402, 0xD040, 0x3002, 0x7400, 0x4840, 0xD182};
/*
   MULU.L       D2,D2:D0
   ADD.W        D0,D0
   MOVE.W       D2,D0
   MOVEQ        #0,D2
   SWAP         D0
   ADDX.L       D2,D0
*/

/* SC craps out if too many words are used in an extern inline */

#  ifdef THINK_C
#   pragma parameter __D0 VFix16Mul(__D0, __D2)
    extern SInt32 VFix16Mul( SInt32, SInt32 ) =
      {0x4C02, 0xC02, 0x6B18, 0xC82, 0, 0x7FFF, 0x6524, 0x6706, 0x70FF, 0xE288,
       0x6026, 0xB0BA, 8, 0x6516, 0x60F2, 0xC82, -1, 0x8000, 0x6E06, 0x7001, 0xE298,
       0x6010, 0x640, 0x7FFF, 0x6002, 0xD040, 0x3002, 0x9582, 0x4840, 0x9082};
/*
   MULU.L       D2,D2:D0
   BMI.S        *+26    ; @3
   CMPI.L       #0x7FFF,D2
   BCS.S        *+38    ; @5
   BEQ.S        *+8     ; @2
@1 MOVEQ        #-1,D0
   LSR.L        #1,D0
   BRA.S        *+40    ; end
@2 CMP.L        *+10,D0
   BCS.S        *+24    ; @5
   BRA.S        *-12    ; @1
@3 CMPI.L       #0xFFFF8000,D2
   BGT.S        *+8     ; @4
   MOVEQ        #1,D0
   ROR.L        #1,D0
   BRA.S        *+18    ; end
@4 ADDI.W       #0x7FFF,D0
   BRA.S        *+4     ; @6
@5 ADD.W        D0,D0
@6 MOVE.W       D2,D0
   SUBX.L       D2,D2
   SWAP         D0
   SUB.L        D2,D0
*/

#   pragma parameter __D0 VFix16Div(__D1, __D2)
    extern SInt32 VFix16Div( SInt32, SInt32 ) =
      {0x2244, 0x2043, 0x2602, 0x6722, 0x6A02, 0x4482, 0xE28A, 0x4A81, 0x6A04, 0x4482,
       0x4083, 0x4841, 0x2001, 0x48C1, 0x4240, 0xD082, 0xD383, 0x4C43, 0xC01, 0x680C,
       0xB781, 0x5AC1, 0x49C1, 0xA81, 0x8000, 0, 0x2608, 0x2809};
/*
   MOVEA.L      D3,A0
   MOVEA.L      D4,A1
   MOVE.L       D2,D4
   BEQ.S        *+36    ; @3
   BPL.S        *+4     ; @1
   NEG.L        D2
@1 LSR.L        #1,D2
   MOVEQ        #0,D3
   TST.L        D1
   BPL.S        *+6     ; @2
   NEG.L        D2
   NEGX.L       D3
@2 SWAP         D1
   MOVE.L       D1,D0
   EXT.L        D1
   CLR.W        D0
   ADD.L        D2,D0
   ADDX.L       D3,D1
   DIVS.L       D4,D1:D0
   BVC.S        *+14    ; end
@3 EOR.L        D4,D1
   SPL          D0
   EXTB.L       D0
   EORI.L       #0x80000000,D0
   MOVE.L       A0,D3
   MOVE.L       A1,D4
*/
#else
   extern SInt32 VFix16Mul( SInt32, SInt32 );
   extern SInt32 VFix16Div( SInt32, SInt32 );
#endif

 /* for performances reasons (more than twice as fast on any M68000 based processor) */
#  ifndef NO_FIXDIV_REPLACEMENT
#   define FixDiv                   VFix16Div
#  endif

#  pragma parameter __D0 NFix16Mul(__D0, __D2)
   extern sfixed NFix16Mul( sfixed, sfixed ) = {0x4C02, 0xC01, 0x3001, 0x4840};
/*
   MULS.L       D2,D1:D0
   MOVE.W       D1,D0
   SWAP         D0
*/

/* WARNING: Passing a denominator of zero to these division inlines _will_ generate
    a divide by zero exception */
#  pragma parameter __D0 NFix16Div(__D0, __D2)
   extern sfixed NFix16Div( sfixed, sfixed ) =
     {0x4840, 0x3200, 0x48C1, 0x4240, 0x4C42, 0xC01};
/*
   SWAP         D0
   MOVE.W       D0,D1
   EXT.L        D1
   CLR.W        D0
   DIVS.L       D2,D1:D0
*/
#  pragma parameter __D0 UFix16Div(__D0, __D2)
   extern ufixed UFix16Div( ufixed, ufixed ) =
     {0x2043, 0x2602, 0xE28B, 0x4281, 0x4840, 0x3200, 0x4240, 0xD083, 0x7600,
      0xD383, 0x4C42, 0x401, 0x2608};
/*  this rounds by adding half the denominator to the 64 bit adjusted numerator
   MOVE.L       D3,A0
   MOVE.L       D2,D3
   LSR.L        #1,D3
   CLR.L        D1
   MOVE.W       D0,D1
   SWAP         D0
   CLR.W        D0
   ADD.L        D3,D0
   MOVEQ        #0,D3
   ADDX.L       D3,D1
   DIVU.L       D2,D1:D0
   MOVE.L       A0,D3
*/

#  pragma parameter __D0 UMulDiv(__D0, __D1, __D2)
   extern SInt32 UMulDiv( SInt32, SInt32, SInt32 ) = {0x4C01, 0x401, 0x4C42, 0x401};
/*
   MULU.L       D1,D1:D0
   DIVU.L       D2,D1:D0
*/

#  pragma parameter __D0 NMulDiv(__D0, __D1, __D2)
   extern SInt32 NMulDiv( SInt32, SInt32, SInt32 ) = {0x4C01, 0xC01, 0x4C42, 0xC01};
/*
   MULS.L       D1,D1:D0
   DIVS.L       D2,D1:D0
*/

   extern void nop(void) = {0x4E71};     /* NOP */
   extern void noop(void) = {0x51FC};    /* TF */

# else /* !env68k */

#  if !defined(__cplusplus)
#   define ClippedU64ToU32(x)       ((-(Int64Hi(x)!=0))|Int64Lo(x))
#  endif

#  ifdef __MRC__ /* the latest version is recommended */

#   define iabs(x)                  __abs(x)
 /* there seems to be no substitute for SGN */

#   define NFract32Mul(x,y)         __mulhw(x,y)
#   define UFract32Mul(x,y)         (__mulhwu(x,y)+((((UInt32)(x))*((UInt32)(y)))>>31))

#   define LZCount32(x)             __cntlzw(x)
#  else
#   define iabs(x)                  ((__cs_itmpa=(x))<0 ? -__cs_itmpa:__cs_itmpa)
#  endif

#  define isgn(x)                   ((int)(__cs_itmpa=(x))<0 ? -1:__cs_itmpa==0 ? 0:1)
#  define ssgn(x)                   ((int)(__cs_itmpa=(x))<(int)(__cs_itmpb=(x)) ? -1:__cs_itmpa==__cs_itmpb ? 0:1)
#  define usgn(x)                   ((unsigned int)(__cs_itmpa=(x))<(unsigned int)(__cs_itmpb=(x)) ? -1:__cs_itmpa==__cs_itmpb ? 0:1)

#  if !COMPILER_HAS_LONGLONG || PREFER_FLOATING_POINT
#   define NFix30Div(x,y)           ((sfract)(((native_double_t)(sfract)(x)*1073741824.0f)/(SInt32)(y)))
#   define NFix16Div(x,y)           ((sfixed)(((native_double_t)(sfixed)(x)*65536.0f)/(SInt32)(y)))
#   define UFix16Div(x,y)           ((ufixed)(((native_double_t)(ufixed)(x)*65536.0f+((ufixed)(y)*0.5f))/(ufixed)(y)))
#   define UMulDiv(x,y,z)           ((UInt32)(((native_double_t)(UInt32)(x)*(y))/(UInt32)(z)))
#   define NMulDiv(x,y,z)           ((SInt32)(((native_double_t)(SInt32)(x)*(y))/(SInt32)(z)))
#  else
#   define NFix30Div(x,y)           ((sfract)((((long long)(sfract)(x))<<30)/(y)))
#   define NFix16Div(x,y)           ((sfixed)((((long long)(sfixed)(x))<<16)/(y)))
#   define UFix16Div(x,y)           ((ufixed)(((((unsigned long long)(ufixed)(x))<<16)+((ufixed)(y)>>1))/(ufixed)(y)))
#   define UMulDiv(x,y,z)           ((UInt32)(((unsigned long long)(UInt32)(x)*(UInt32)(y))/(UInt32)(z)))
#   define NMulDiv(x,y,z)           ((SInt32)(((long long)(SInt32)(x)*(SInt32)(y))/(SInt32)(z)))
#  endif

#  if !COMPILER_HAS_LONGLONG /* && !defined(powerc) -- && PREFER_FLOATING_POINT)*/
#   ifndef NFract32Mul
#    define NFract32Mul(x,y)        ((SInt32)((((SInt32)(y))*(native_double_t)(SInt32)(x))/4294967296.0f))
#   endif
#   ifndef UFract32Mul
#    define UFract32Mul(x,y)        ((UInt32)((((native_double_t)(UInt32)(x)*(UInt32)y)+2147483648.0f)/4294967296.0f))
#   endif
#   define NFix30Mul(x,y)           ((sfract)(((y)*(native_double_t)(sfract)(x))/1073741824.0f))
#   define UFix16Mul(x,y)           ((ufixed)(((y)*(native_double_t)((ufixed)(x))+32768.0f)/65536.0f))
#   define NFix16Mul(x,y)           ((sfixed)(((y)*(native_double_t)(sfixed)(x))/65536.0f))
#  else
#   ifndef NFract32Mul
#    define NFract32Mul(x,y)        ((sfixed)((((SInt32)(y))*(long long)(sfixed)(x))>>32))
#   endif
#   ifndef UFract32Mul
#    define UFract32Mul(x,y)        ((UInt32)((((unsigned long long)(UInt32)(x)*(UInt32)y)+2147483648ULL)>>32))
#   endif
#   define NFix30Mul(x,y)           ((sfract)((((SInt32)(y))*(long long)(sfract)(x))>>30))
#   define UFix16Mul(x,y)           ((ufixed)((((ufixed)(y))*(unsigned long long)(ufixed)(x)+32768ULL)>>16))
#   define NFix16Mul(x,y)           ((sfixed)((((SInt32)(y))*(long long)(sfixed)(x))>>16))
#  endif

#  ifndef LZCount32
    EXTERN_API_C( int ) LZCount32( UInt32 x );
#   if defined(__cplusplus)
      inline int LZCount32(UInt32 x) {
         int i = 32;

         do{if(x&0x80000000) break; x += x;}while(--i);
         return 32-i;
      }
#   endif
#  endif

#  if COMPILER_HAS_LONGLONG
#   ifndef __DEF_SWAP
#    undef SWAP
#    define SWAP(x,y)               do{if(sizeof(x)==1) {UInt8 _swaptmp=x; x=y; \
      y=_swaptmp;} else if(sizeof(x)==2) {SInt16 _swaptmp=x; x=y; y=_swaptmp;} \
      else if(sizeof(x)==4) {SInt32 _swaptmp=x; x=y; y=_swaptmp;} \
      else if(sizeof(x)==8) {UInt64 _swaptmp=x; x=y; y=_swaptmp;}}while(0)
#   endif

#   define kZero64                  0
#   define kInt64U32(x)             (x)
#   define kInt64S32(x)             (x)

#   define SetInt64(x,y,z)          do{(x)=((long long)(y)<<32)+(UInt32)(z);}while(0)
#   define LSetInt64(x,y)           do{(x)=(long int)(y);}while(0)
#   define XSetInt64(x,y)           do{(x)=(0.0+(y));}while(0)

#   define SAdd64(x,y)              do{(x)+=(SInt64)(y);}while(0)
#   define UAdd64(x,y)              do{(x)+=(UInt64)(y);}while(0)
#   define SSub64(x,y)              do{(x)-=(SInt64)(y);}while(0)
#   define USub64(x,y)              do{(x)-=(UInt64)(y);}while(0)

#   define SNeg64(x)                do{(x)=-(SInt64)(x);}while(0)
#   define UNeg64(x)                do{(x)=-(UInt64)(x);}while(0)

#   define Add64Int(x,y,z)          do{(x)+=((y)<<32)+(UInt32)(z);}while(0)
#   define Sub64Int(x,y,z)          do{(x)-=((y)<<32)+(UInt32)(z);}while(0)
#   define Add64S32(x,y)            do{(x)+=(SInt32)(y);}while(0)
#   define Add64U32(x,y)            do{(x)+=(UInt32)(y);}while(0)
#   define Sub64S32(x,y)            do{(x)-=(SInt32)(y);}while(0)
#   define Sub64U32(x,y)            do{(x)-=(UInt32)(y);}while(0)

#   define LShiftU64(x,y)           do{(x)=(UInt64)(x)<<(y);}while(0)
#   define LShiftS64(x,y)           do{(x)=(SInt64)(x)<<(y);}while(0)
#   define RShiftU64(x,y)           do{(x)=(UInt64)(x)>>(y);}while(0)
#   define RShiftS64(x,y)           do{(x)=(SInt64)(x)>>(y);}while(0)
#   define DoubleU64(x)             do{(x)+=(x);}while(0)
#   define DoubleS64(x)             do{(x)+=(x);}while(0)
#   define HalfU64(x)               do{(x)=(UInt64)(x)>>1;}while(0)
#   define HalfS64(x)               do{if((SInt64)(x)<0) ++(x); (x)>>=1;}while(0)

#   ifdef __cplusplus
     inline void UMul64U32(UInt64 &x, UInt32 y) {(x)*=(y);}
     inline void SMul64U32(SInt64 &x, SInt32 y) {(x)*=(y);}
     inline void UMul64S32(UInt64 &x, UInt32 y) {(x)*=(y);}
     inline void SMul64S32(SInt64 &x, SInt32 y) {(x)*=(y);}

     inline void UDiv64U32(UInt64 &x, UInt32 y) {(x)/=(y);}
     inline void UDiv64U32R(UInt64 &x, UInt32 y, UInt32 &z) {(z)=(x); (x)/=(y); (z)-=(x)*(y);}
     inline void SDiv64S32(SInt64 &x, SInt32 y) {(x)/=(y);}
     inline void SDiv64S32R(SInt64 &x, SInt32 y, SInt32 &z) {(z)=(x); (x)/=(y); (z)-=(x)*(y);}

     inline UInt32 _UDiv64U32R(UInt64 *x, UInt32 y) { UInt32 _tmp=*(x); *(x)/=(y); return _tmp-*(x)*(y);}
     inline SInt32 _SDiv64S32R(SInt64 *x, SInt32 y) { SInt32 _tmp=*(x); *(x)/=(y); return _tmp-*(x)*(y);}

     inline bool U64Gt(UInt64 x, UInt64 y) {return (x)>(y);}
     inline bool U64Ge(UInt64 x, UInt64 y) {return (x)>=(y);}
     inline bool S64Gt(SInt64 x, SInt64 y) {return (x)>(y);}
     inline bool S64Ge(SInt64 x, SInt64 y) {return (x)>=(y);}

     inline bool U64LtU32(UInt64 x, UInt32 y) {return (x)<(y);}
     inline bool S64LtU32(SInt64 x, UInt32 y) {return (x)<(y);}
     inline bool U64LeU32(UInt64 x, UInt32 y) {return (x)<=(y);}
     inline bool S64LeU32(SInt64 x, UInt32 y) {return (x)<=(y);}
     inline bool U64GtU32(UInt64 x, UInt32 y) {return (x)>(y);}
     inline bool S64GtU32(SInt64 x, UInt32 y) {return (x)>(y);}
     inline bool U64GeU32(UInt64 x, UInt32 y) {return (x)>=(y);}
     inline bool S64GeU32(SInt64 x, UInt32 y) {return (x)>=(y);}

     inline void UMul32(UInt64 &x, UInt32 y, UInt32 z) {(x)=(y)*(z);}
     inline void SMul32(SInt64 &x, SInt32 y, SInt32 z) {(x)=(y)*(z);}

     inline UInt32 UDiv32(UInt64 &x, UInt32 y) {return (x)/(y);}
     inline UInt32 UMod32(UInt64 &x, UInt32 y) {return (x)%(y);}
     inline UInt32 SDiv32(SInt64 &x, SInt32 y) {return (x)/(y);}
     inline UInt32 SMod32(SInt64 &x, SInt32 y) {return (x)%(y);}
#   else
#    define UMul64U32(x,y)          do{(x)=(UInt64)(x)*(UInt32)(y);}while(0)
#    define SMul64U32(x,y)          do{(x)=(SInt64)(x)*(UInt32)(y);}while(0)
#    define UMul64S32(x,y)          do{(x)=(UInt64)(x)*(SInt32)(y);}while(0)
#    define SMul64S32(x,y)          do{(x)=(SInt64)(x)*(SInt32)(y);}while(0)

#    define UDiv64U32(x,y)          do{(x)/=(UInt32)(y);}while(0)
#    define UDiv64U32R(x,y,z)       do{(z)=(UInt32)(x); (x)/=(SInt32)(y); (z)-=(UInt32)((x)*(y));}while(0)
#    define SDiv64S32(x,y)          do{(x)/=(SInt32)(y);}while(0)
#    define SDiv64S32R(x,y,z)       do{(z)=(UInt32)(x); (x)/=(SInt32)(y); (z)-=(UInt32)((x)*(y));}while(0)

 /* special implementation; not recommended -- use xDiv64x32R or target C++ instead */
#    define _UDiv64U32R(x,y)        (0+(*(UInt32*)&temporary=*(UInt64*)(x),*(UInt64*)(x)/=(UInt32)(y),*(UInt32*)&temporary-*(UInt64*)(x)*(y)))
#    define _SDiv64S32R(x,y)        (0+(*(SInt32*)&temporary=*(UInt64*)(x),*(UInt64*)(x)/=(SInt32)(y),*(SInt32*)&temporary-*(UInt64*)(x)*(y)))

#    define U64Gt(x,y)              ((UInt64)(x)>(UInt64)(y))
#    define U64Ge(x,y)              ((UInt64)(x)>=(UInt64)(y))
#    define S64Gt(x,y)              ((SInt64)(x)>(SInt64)(y))
#    define S64Ge(x,y)              ((SInt64)(x)>=(SInt64)(y))

#    define U64LtU32(x,y)           (((UInt64)(x))<(UInt32)(y))
#    define S64LtU32(x,y)           (((SInt64)(x))<(UInt32)(y))
#    define U64LeU32(x,y)           (((UInt64)(x))<=(UInt32)(y))
#    define S64LeU32(x,y)           (((SInt64)(x))<=(UInt32)(y))
#    define U64GtU32(x,y)           (((UInt64)(x))>(UInt32)(y))
#    define S64GtU32(x,y)           (((SInt64)(x))>(UInt32)(y))
#    define U64GeU32(x,y)           (((UInt64)(x))>=(UInt32)(y))
#    define S64GeU32(x,y)           (((SInt64)(x))>=(UInt32)(y))

#    define UMul32(x,y,z)           do{(x)=((UInt64)(UInt32)(y))*(UInt32)(z);}while(0)
#    define SMul32(x,y,z)           do{(x)=((SInt64)(SInt32)(y))*(SInt32)(z);}while(0)

#    define UDiv32(x,y)             ((UInt32)(((UInt64)(x))/(UInt32)(y)))
#    define UMod32(x,y)             ((UInt32)(((UInt64)(x))%(UInt32)(y)))
#    define SDiv32(x,y)             ((SInt32)(((SInt64)(x))/(SInt32)(y)))
#    define SMod32(x,y)             ((SInt32)(((SInt64)(x))%(SInt32)(y)))
#   endif

#   define Int64Eq(x,y)             (((UInt64)(x))==(UInt64)(y))
#   define Int64EqU32(x,y)          (((SInt64)(x))==(UInt32)(y))

#   define Int64Not(x)              do{(x)=!(UInt64)(x);}while(0)
#   define Int64BNot(x)             do{(x)=~(UInt64)(x);}while(0)
#   define Int64BOr(x,y,z)          do{(x)|=(((UInt64)(y))<<32)|(UInt32)(z);}while(0)
#   define Int64BAnd(x,y,z)         do{(x)&=(((UInt64)(y))<<32)|(UInt32)(z);}while(0)
#   define Int64BExOr(x,y,z)        do{(x)^=(((UInt64)(y))<<32)|(UInt32)(z);}while(0)
#   define BitOr64(x,y)             do{(x)|=(UInt64)(y);}while(0)
#   define BitAnd64(x,y)            do{(x)&=(UInt64)(y);}while(0)
#   define BitExOr64(x,y)           do{(x)^=(UInt64)(y);}while(0)

#   define Int64Z(x)                (((UInt64)(x))==0)
#   define Int64N(x)                (((SInt64)(x))<0)
#   define Int64P(x)                (((SInt64)(x))>0)

#   define Int64Hi(x)               ((x)>>32)
#   define Int64Lo(x)               ((x)&0xFFFFFFFFUL)
#   define Int64ToL(x)              (long int)(0+(x))
#   define Int64ToX(x)              ((long double)(0+(x)))
#  endif

#  define SET_FLAG8(x,y,z)          _SET_FLAG8(x,y,z)
#  define SET_FLAG16(x,y,z)         _SET_FLAG16(x,y,z)
#  define SET_FLAG32(x,y,z)         _SET_FLAG32(x,y,z)

#  define nop()                     do{ }while(0)
#  define noop()                    do{ }while(0)
# endif

# define _SET_FLAG8(x,y,z)          do{if(z) x|=(1<<(y)); else x&=~(1<<(y));}while(0)
# define _SET_FLAG16(x,y,z)         _SET_FLAG8(x,y,z)
# define _SET_FLAG32(x,y,z)         do{if(z) x|=(1L<<(y)); else x&=~(1L<<(y));}while(0)

# define U64Lt(x,y)                 !U64Ge(x,y)
# define U64Le(x,y)                 !U64Gt(x,y)
# define S64Lt(x,y)                 !S64Ge(x,y)
# define S64Le(x,y)                 !S64Gt(x,y)

# define GET_ARRAY_INDEX(x,y)       (((unsigned long int)(y)-(unsigned long int)(x))/sizeof((x)[0]))
# define GET_ARRAY_COUNT(x)         ((unsigned long int)sizeof(x)/sizeof((x)[0]))

/* this comes standard with THINK C... */
# ifndef OFFSET
#  define OFFSET(t,w)               ((int)&((t*)0)->w)
# endif

# ifndef __MAKE_LITERAL_STR
#  define __MAKE_LITERAL_STR(x)     #x
# endif

/*
  temporary and empty variables; the system specific CSeries library supplies
 "empty"; it is nothing but zeros and should also work as an empty C-string;

  both can be casted as anything
*/
   extern char          temporary[];
   extern const UInt8   empty[];

/* these are reserved by the macros in this file (you shouldn't use them) */
   extern int           __cs_itmpa, __cs_itmpb, __cs_itmpc;

/*  the video constants and macros have moved to cs_graphics.h */

/*  universal Graphics/Text constants */

   enum
   {
    /* clipping methods */
      _cs_clip_completely           = 0, /* remains empty if it goes beyond maximum */
      _cs_clip_left_to_right_char   = 1, /* only valid for text */
      _cs_clip_right_to_left_char   = 2, /* only valid for text */
      _cs_clip_at_middle_char       = 3, /* only valid for text */
      _cs_clip_to_first_object      = 4, /* only first (or empty it's too long) if beyond boundary */
      _cs_clip_left_to_right        = 5,
      _cs_clip_right_to_left        = 6,
      _cs_clip_at_middle_object     = 7,

    /* text-only alternative names */
      _cs_clip_to_first_word        = _cs_clip_to_first_object,
      _cs_clip_left_to_right_word   = _cs_clip_left_to_right,
      _cs_clip_right_to_left_word   = _cs_clip_right_to_left,
      _cs_clip_at_middle_word       = _cs_clip_at_middle_object
   };

/*  Debugging */

   /* the debugging string constants are no longer supported */

/* any negative value passed for "level" means terminate the program
    any positive value means to pause (so zero would mean to continue)
*/
   EXTERN_API_C(void) _debugMsg(const char*, const char*, int lineNumber, int level);


/* program-implemented malloc prototypes */

  /* "transitional malloc"; custom to the project (and must be supplied by it) */
   EXTERN_API_C(void*) tmalloc(size_t size);

/***

   Define "FORCE_CONDITION_ENCODING" to a non-zero value if you want the compiler to
   encode the expression regardless of whether DEBUG is defined.

   If "FORCE_CONDITION_ENCODING" is not defined, then it is set to 1 as a precaution.

***

  Revision Note:
     As of version 0.3, the _debugMsg macros pass their parameters differently.
   You should recompile any source files that use them.

***/

# ifndef FORCE_CONDITION_ENCODING
#  define FORCE_CONDITION_ENCODING      1
# endif

/* assertion */
# ifndef assert
#  if DEBUG
#   define assert(x)                do{if(!(x)) _debugMsg(#x,__source__,__LINE__,1);}while(0)
#  else
#   if FORCE_CONDITION_ENCODING
#    define assert(x)               do{if(!(x)) __cs_itmpa=0;}while(0)
#   else
#    define assert(x)
#   endif
#  endif

#  if DEBUG
#   define vassert(x,y)             do{if(!(x)) _debugMsg(y,__source__,__LINE__,1);}while(0)
#  else
#   if FORCE_CONDITION_ENCODING
#    define vassert(x,y)            do{if(!(x)) __cs_itmpa=0;}while(0)
#   else
#    define vassert(x,y)
#   endif
#  endif

/* "warning" */
#  if DEBUG
#   define warn(x)                  do{if(!(x)) _debugMsg(#x,__source__,__LINE__,0);}while(0)
#  else
#   if FORCE_CONDITION_ENCODING
#    define warn(x)                 do{if(!(x)) __cs_itmpa=0;}while(0)
#   else
#    define warn(x)
#   endif
#  endif

#  if DEBUG
#   define vwarn(x,y)               do{if(!(x)) _debugMsg(y,__source__,__LINE__,0);}while(0)
#  else
#   if FORCE_CONDITION_ENCODING
#    define vwarn(x,y)              do{if(!(x)) __cs_itmpa=0;}while(0)
#   else
#    define vwarn(x,y)
#   endif
#  endif

/* "pause" program */
#  if DEBUG
#   define pause()                  _debugMsg(NULL,__source__,__LINE__,1)
#  else
#   define pause()
#  endif

#  if DEBUG
#   define vpause(x)                _debugMsg(x,__source__,__LINE__,1)
#  else
#   define vpause(x)
#  endif

/* "halt": exit the program - encoded regardless of DEBUG */
#  define halt()                    _debugMsg(NULL,__source__,__LINE__,-1)
#  define vhalt(x)                  _debugMsg(x,__source__,__LINE__,-1)
# endif

# ifdef __cplusplus
   }
# endif

# if PRAGMA_STRUCT_ALIGN
#  pragma options align=reset
# elif PRAGMA_STRUCT_PACKPUSH
#  pragma pack(pop)
# elif PRAGMA_STRUCT_PACK
#  pragma pack()
# endif

# ifdef PRAGMA_IMPORT_OFF
#  pragma import off
# elif PRAGMA_IMPORT
#  pragma import reset
# endif

#endif  /* __MY_CSERIES__ */
