/*
 byte_swapping.c
  MyCSeries source file for endian byte swapping

  Version 0.4.1

  Created: October 28, 2004
  Revised: June 24, 2007

 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 *  This source code is offered for use in the public domain.  You may use,
 * modify or distribute it freely.
 *
 *  This code is distributed in the hope that it will be useful but WITHOUT
 * ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY DISCLAIMED.
 *  This includes but is not limited to warranties of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <cseries.h>

#include <Debugging.h>

#include "cs_byte_swapping.h"

/* private prototypes */

static UInt8 *_byte_swap_data(void *d, long size, long count, const UInt8 *f);

/* code */

void *reverse_bytes(long c, void *_d)
{
	UInt8			*s, *d = c+(UInt8*)_d, t;

	if ((s=(UInt8*)d)==(UInt8*)NULL || s==(UInt8*)NONE)
		return NULL;
	if (c<=0)
		return _d;
	do
	{
		t = *--d;
		*d = *s;
		*s++ = t;
	} while(--c);

	return (void*)s;
}

void byte_swap_memory(UInt8 *b,
	UInt8 w,
	void *d,
	UInt8 t,
	long count)
{
	UInt8			*s = (UInt8*)d;
	int				entry;
	Boolean			t1, t2, t3, t4;

	if (b!=(UInt8*)NULL)
		if ((*b!=0)==(w!=0))
			return;
	if (count<=0 || (entry=t)<=0x88) /* these are invalid here */
		return;
	switch(entry)
	{
	case -1:
		break; /* swap with self (does nothing) */
	case _2byte:
		/* unrolled once */
		if (count&1)
		{
			t1 = *s; *s++ = s[1]; *s++ = t1;
		}
		if ((count>>=1)==0)
			return;
		do
		{
			t1 = *s; *s++ = s[1]; *s++ = t1;
			t1 = *s; *s++ = s[1]; *s++ = t1;
		} while(--count);
		break;

	case _4byte:
		do
		{
			t1 = *s; *s++ = s[3];
			t2 = *s; *s++ = s[2];

			*s++ = t2;
			*s++ = t1;
		} while(--count);
		break;

	case _8byte:
		do
		{
			t1 = *s; *s++ = s[7];
			t2 = *s; *s++ = s[6];
			t3 = *s; *s++ = s[5];
			t4 = *s; *s++ = s[4];

			*s++ = t4;
			*s++ = t3;
			*s++ = t2;
			*s++ = t1;
		} while(--count);
		break;

	default:
		do
		{
			s = (UInt8*)reverse_bytes(-(SInt8)t,(UInt8*)s);
		} while(count-=t);
		break;
	}

	if (b!=(UInt8*)NULL)
		*b = w;
}

int byte_swap_data(UInt8 *b,
	UInt8 w,
	void *s,
	long z,
	long c,
	const UInt8 *f)
{
	if (b!=(UInt8*)NULL)
		if ((*b!=0)==(w!=0))
			return 2;

	if ((s=_byte_swap_data(s,z,c,f))==NULL)
		return 0;

	if (b!=(UInt8*)NULL)
		*b = w;

	return 1;
}

int byte_swap_array(UInt8 *b,
	UInt8 w,
	void *s,
	long z,
	long r,
	const UInt8 *f)
{
	if (b!=(UInt8*)NULL)
		if ((*b!=0)==(w!=0))
			return 2;
	if (r>0) do
	{
		if ((s=_byte_swap_data(s,z,LONG_MAX,f))==NULL)
			return 0;
	} while(--r);

	if (b!=(UInt8*)NULL)
		*b = w;

	return 1;
}

/* private code */

static UInt8 *_byte_swap_data(void *d, long size, long count, const UInt8 *f)
{
	UInt8			*s;
	long			temp;
	UInt8			t1, t2, t3, t4;

	if ((s=(UInt8*)d)==(UInt8*)NULL || (count<=0) || (size<=0))
		goto end;
	do
	{
		if ((temp=*f++)==_end)
			goto end; /* 0 */
		switch(temp)
		{
		case -1:
			break; /* swap with self (does nothing) */
		case _2byte:
			if (size>=2)
			{
				size -= 2;
				t1 = *s; *s++ = s[1]; *s++ = t1;
			} else
				return (UInt8*)NULL;
			break;

		case _4byte:
			if (size>=4)
			{
				t1 = *s; *s++ = s[3];
				t2 = *s; *s++ = s[2];

				*s++ = t2;
				*s++ = t1;
			} else
				return (UInt8*)NULL;
			break;

		case _8byte:
			if (size>=8)
			{
				t1 = *s; *s++ = s[7];
				t2 = *s; *s++ = s[6];
				t3 = *s; *s++ = s[5];
				t4 = *s; *s++ = s[4];

				*s++ = t4;
				*s++ = t3;
				*s++ = t2;
				*s++ = t1;
			} else
				return (UInt8*)NULL;
			break;
		case _array8:
			temp = *s++;
			goto do_array;
		case _array16:
#if TARGET_RT_LITTLE_ENDIAN
			temp = *s++<<8; temp |= *s++;
#else
			temp = *(UInt16*)s; s += 2;
#endif
			goto do_array;
		case _array32:
#if TARGET_RT_LITTLE_ENDIAN
			temp = *s++<<24; temp |= *s++<<16; temp |= *s++<<8; temp |= *s++;
#else
			temp = *(UInt32*)s; s += 4;
#endif
		do_array:
			do
			{
				_byte_swap_data(s,size,count-(long)d+(long)s,f);
			} while(--temp);
			break;
		case _entry32:
#if TARGET_RT_LITTLE_ENDIAN
			temp = *s++<<24; temp |= *s++<<16; temp |= *s++<<8; temp |= *s++;
#else
			temp = *(UInt32*)s; s += 4;
#endif

		default:
			if (temp<0x7F)
			{
				s += temp;
				break;
			} else if (temp<0xFF)
				temp = (SInt8)temp;
			if (size-=(temp=-temp)<0)
				return (UInt8*)NULL;
			s = (UInt8*)reverse_bytes(-temp,(UInt8*)s);
			break;
		}
	} while(--count);
end:
	return s;
}
