/*
 files_macintosh.c
  Source file for cs_files on MacOS platform

  Version 0.4.1

  Created: January 21, 2006
  Revised: June 24, 2007

 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 *  This source code is offered for use in the public domain.  You may use,
 * modify or distribute it freely.
 *
 *  This code is distributed in the hope that it will be useful but WITHOUT
 * ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY DISCLAIMED.
 *  This includes but is not limited to warranties of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <macintosh_cseries.h> /* all of the "cs_" files are "included" */

#include <Gestalt.h>

/* constants */

#define DEFAULT_FILE_COPY_BUFFER_SIZE	(128*KILO)

#define HFSPLUS_COMPATIBLE_BITS			((1<<gestaltFSSupports2TBVols)|(1<<gestaltFSSupportsHFSPlusVols))

/* notice that with the ".FRK" extension, this is exactly the same name
  that PC Exchange uses */
char			cs_resource_fork_name[] = "RESOURCE";

/* globals */

int				file_system_info = 0; /* zero means "not initialized" */

static Str255			name_temporary = "\p";
static HFSUniStr255		unistr_temporary = {0};

static HParamBlockRec	hpb_temporary;
static ParamBlockRec	pb_temporary;
static FSForkIOParam	fsf_temporary;
static FSRefParam		fsr_temporary;

/* private prototypes */

/* now uses hpb_temporary */
static int allocate_non_native_directory_fork(char *name);
static OSErr copy_fork_to_non_native_file(HParmBlkPtr hpb, const char *fork, SInt64 eof);

static void evaluate_file_system(void);
static int cs_file_to_fsref(const cs_filespec *src, FSRef *dst, UInt8 permission);

static int cstring_to_unistring(char *str, HFSUniStr255 *ustr);

/* code */

OSErr get_my_fsspec(FSSpec *file)
{
	OSErr			err;

	ProcessSerialNumber	psn;
	ProcessInfoRec		processInfo;

	processInfo.processInfoLength = sizeof(ProcessInfoRec);
	processInfo.processName = (UInt8*)NULL;
	if ((err=GetCurrentProcess(&psn))!=noErr)
		return err;
	processInfo.processAppSpec = (struct FSSpec*)file;

	/* jeez, all o'that wasted info... */
	return GetProcessInformation(&psn,&processInfo);
}

int fs_locate_file_by_strlist(cs_filespec *outFile,
	short strID, int index, int range,
	short pathsStrID)
{
	FSSpec			file;
	Str255			astr;
	CInfoPBRec		cpb;
	HParamBlockRec	hpb;
	SInt16			count, err;

	if (range<=0)
		return kCS_parameter_error;

	/* handle NONE pointers as a precaution */
	if (outFile!=(cs_filespec*)NULL && outFile!=(cs_filespec*)NONE)
	{
/* someday, I'll actually implement this to handle other encodings */
		if (outFile->encoding!=_text_encoding_MacRoman)
			return kTS_encoding_not_found;
	}

	get_my_fsspec(&file);
	do
	{
		getpstr(astr,strID++,index);
		cpb.hFileInfo.ioCompletion = (IOCompletionUPP)NULL;
		hpb.ioParam.ioCompletion = (IOCompletionUPP)NULL;
		cpb.hFileInfo.ioVRefNum = file.vRefNum, cpb.hFileInfo.ioNamePtr = astr,
			cpb.hFileInfo.ioDirID = file.parID, *(SInt32*)&cpb.hFileInfo.ioFDirIndex = 0;
		if ((err=PBGetCatInfoSync(&cpb))==noErr)
			goto found;	/* found it */
	
		/* not in the app's root directory, must use the paths strings */
		if ((count=countstrlist(pathsStrID))>0)
		{
			do
			{
				/*
					in order to get the dir. and volume,
					we need get the full specs on it
				*/
				hpb.ioParam.ioMisc = (Ptr)&file, hpb.ioParam.ioVRefNum = 0,
					hpb.ioParam.ioNamePtr = getpstr(astr,pathsStrID,index);
				err = PBMakeFSSpecSync(&hpb);
	/*aptr = (SInt16*)(*(UInt8*)aptr+(Ptr)aptr);*/
				if (err) continue;							/* directory non-existant */
	
				/* get inside the directory
					remember, we want to save the name of the file */
				cpb.dirInfo.ioDrDirID = 0, *(SInt32*)&cpb.dirInfo.ioFDirIndex = 0,
					cpb.dirInfo.ioNamePtr = file.name;
				/* if getting the catalog info for the directory doesn't work,
					that would usually mean it the volume was taken offline */
				if ((err=PBGetCatInfoSync(&cpb))!=noErr) continue;
				err = -43;		/* not found */
				if (!cpb.dirInfo.ioDrNmFls) continue;		/* the directory's empty */
	
				/* go directly to the file (the dirID field is already set) */
				cpb.hFileInfo.ioVRefNum = file.vRefNum,
					cpb.hFileInfo.ioNamePtr = astr;
				if ((err=PBGetCatInfoSync(&cpb))==noErr)
					break; /* found it */
			} while(--count);
		}
	} while(--range);
	/* we don't want to return that a volume isn't found 'cause we're searching
	  for a file (and we certainly exist on a volume)
	*/
	if (err==nsvErr)
		err = fnfErr;

	/* should be "file not found" */
	if (err) return translate_platform_error(err,_platform_MacOS);
found:
	if (outFile!=(cs_filespec*)NULL && outFile!=(cs_filespec*)-1)
	{
		outFile->directory = file.parID;
		outFile->volume = file.vRefNum;
		if (outFile->name)
			BlockMoveData(&astr[1],outFile->name,outFile->length=astr[0]);
	}
	return kCS_no_error;
}

int build_platform_filespec(cs_filespec *inFile,
	platform_filespec *outFile)
{
	if (inFile->encoding!=_text_encoding_MacRoman)
		return kTS_encoding_cannot_be_used;

	if (inFile->length>31)
		return kFS_name_too_long;

	/*outFile->vRefNum = inFile->volume;*/
	/*outFile->parID = inFile->directory;*/

	if (inFile->name!=(char*)NULL)
	{
		name_temporary[0] = inFile->length;
		BlockMoveData(inFile->name,&name_temporary[1],inFile->length);
	} else
		name_temporary[0] = 0;

	return translate_platform_error(FSMakeFSSpec(inFile->volume,
		inFile->directory,name_temporary,outFile),_platform_MacOS);
}

int fs_create(cs_filespec *desiredFile,
    UInt32 flags,
    UInt32 type,
    UInt32 creator)
{
	int				result = 0;

	if (desiredFile!=(cs_filespec*)NULL && desiredFile!=(cs_filespec*)NONE)
		return kCS_parameter_error;

/* not yet handled */
	if (desiredFile->encoding!=_text_encoding_MacRoman)
		return kTS_encoding_not_found;

	if (desiredFile->length>31)
		return kFS_name_too_long;

	hpb_temporary.fileParam.ioFVersNum = 0; /* must be cleared as far as I can tell */
	hpb_temporary.fileParam.ioDirID = desiredFile->directory;
	hpb_temporary.fileParam.ioVRefNum = desiredFile->volume;

	if (desiredFile->name!=(char*)NULL)
	{
		name_temporary[0] = desiredFile->length;
		BlockMoveData(desiredFile->name,&name_temporary[1],desiredFile->length);
	} else
		name_temporary[0] = 0;

	hpb_temporary.fileParam.ioNamePtr = name_temporary;

	if ((flags&_create_non_native_forks) || desiredFile->fork!=(char*)NULL)
	{
		SInt32		old_directory;
		char		*fork_name = (char*)NULL;

		/* create the resource fork directory by default */
		if (flags&_create_non_native_forks)
		{
			if ((fork_name=desiredFile->fork)==(char*)NULL || (*fork_name)==0)
				fork_name = cs_resource_fork_name;

			if ((result=allocate_non_native_directory_fork(fork_name))<0)
				return result;

			if ((fork_name=desiredFile->fork)==(char*)NULL || (*fork_name)==0)
				fork_name = (char*)NULL;
		} else if (*(desiredFile->fork)!=0) /* shouldn't be NULL in this case */
		{
			if (file_system_info==0)
				evaluate_file_system();
			/* for the HFS and certain other related file systems, the resource
			  fork always exists (even though it might be empty); only create
			  the non-native fork if the volume is not HFS+ or the extended
			  file manager is not available (Mac OS 9 and later) */
			if (strcmp(desiredFile->fork,cs_resource_fork_name))
			{
				if (file_system_info==0) evaluate_file_system();
				if (file_system_info<=0)
				{	/* either HFS+ APIs aren't supported or init. failed */
					fork_name = desiredFile->fork;

					if ((result=allocate_non_native_directory_fork(fork_name))<0)
						return result;
				} else
				{
					/* if the name is eight chars, add "_FORK" to the name
					  so "RESOURCE" comes out "RESOURCE_FORK" (recognized as the
					  resource fork in the HFS+ extended file system) */
					
				}
			}
		} else fork_name = (char*)NULL;

		if (fork_name!=(char*)NULL)
		{
			old_directory = hpb_temporary.fileParam.ioDirID;
			hpb_temporary.fileParam.ioDirID = result;

			if ((result=PBHCreateSync(&hpb_temporary))!=0)
			{
				/* ignore errors that over creating a file that already exists
				  for the non-native fork */
				if (result!=-48) /* dupFNErr */
					return translate_platform_error(result,_platform_MacOS);
			}
			hpb_temporary.fileParam.ioDirID = old_directory;

			/* fall-in to make sure the data fork (or the main file) exists */
		}
	}

	if ((result=PBHCreateSync(&hpb_temporary))!=0)
	{
		/* ignore the duplicate file name error if we're just creating a fork */
		if (desiredFile->fork==(char*)NULL || result!=-48) /* dupFNErr */
			return translate_platform_error(result,_platform_MacOS);
	} else if (type!=0)
	{
		if ((result=PBHGetFInfoSync(&hpb_temporary))!=noErr)
			return translate_platform_error(result,_platform_MacOS);

		hpb_temporary.fileParam.ioFlFndrInfo.fdType = type;
		hpb_temporary.fileParam.ioFlFndrInfo.fdCreator = creator;

		return translate_platform_error(PBHSetFInfoSync(&hpb_temporary),_platform_MacOS);
	}
	return kCS_no_error;
}

int fs_push_out_forks(cs_filespec *inFile)
{
	SInt64			wtemp;
	OSErr			err;

	if (file_system_info==0) evaluate_file_system();

	/* create all non-native allocating directories and move the fork data
	  to the representing files by copying and emptying the native forks
	  (except for the data fork) */

	if (file_system_info<=0)
	{ /* only the data and resource forks are native */
		hpb_temporary.ioParam.ioPermssn = fsWrPerm;
		hpb_temporary.fileParam.ioFVersNum = 0;
		hpb_temporary.fileParam.ioVRefNum = inFile->volume;
		hpb_temporary.fileParam.ioDirID = inFile->directory;
		
		if (inFile->name!=(char*)NULL)
		{
			name_temporary[0] = inFile->length;
			BlockMoveData(inFile->name,&name_temporary[1],inFile->length);
		} else
			name_temporary[0] = 0;
		hpb_temporary.fileParam.ioNamePtr = name_temporary;

		if ((err=PBHOpenRFSync(&hpb_temporary))!=noErr)
		{
			if (err==-49) /* opWrErr */
				return kFS_fork_busy;
			return translate_platform_error(err,_platform_MacOS);
		}

		/* GetEOF can't do crap here as far as wide positioning, so we will
		  have to use SetFPos and temporarily set the position to the end
		  to get the file's logical end */
		if (file_system_info<-1) /* worst case senerio (ancient system) */
			hpb_temporary.ioParam.ioPosMode = fsFromLEOF;
		else
			hpb_temporary.ioParam.ioPosMode = fsFromLEOF | kUseWidePositioning;
		if ((err=PBSetFPosSync((ParmBlkPtr)&hpb_temporary))!=noErr)
			return translate_platform_error(err,_platform_MacOS);

		if (file_system_info<-1)
			SetInt64(wtemp,0,hpb_temporary.ioParam.ioPosOffset);
		else
			wtemp = *(SInt64*)&((*(XIOParam*)&hpb_temporary).ioWPosOffset);

		if ((err=copy_fork_to_non_native_file(&hpb_temporary,
				cs_resource_fork_name,wtemp))!=noErr)
			return translate_platform_error(err,_platform_MacOS);

		/* and empty the resource fork */
		hpb_temporary.ioParam.ioMisc = (Ptr)0;
		if ((err=PBSetEOFSync((ParmBlkPtr)&hpb_temporary))!=noErr)
			return translate_platform_error(err,_platform_MacOS);

		/* and because the non-native forks are already non-native, we are done! */

	} else
	{
		CatPositionRec	catPos;
		FSRef			fs_reference;
		SInt64			eof;

		if ((err=cs_file_to_fsref(inFile,&fs_reference,fsWrPerm))!=0)
			return err;

		memset((void*)&catPos,0,sizeof(catPos));
		err = FSIterateForks(&fs_reference,&catPos,&unistr_temporary,&eof,(UInt64*)NULL);
		
	}
	return kCS_no_error;
}

int fs_open_fork(cs_filespec *srcFile,
    UInt32 flags)
{
	char			*fork_name = (char*)NULL;
	OSErr			err;
	UInt8			permissions = (flags&_open_write_permission) ? fsWrPerm : fsRdPerm;

	if (file_system_info==0)
		evaluate_file_system();
	if (file_system_info<=0)
	{ /* only the data and resource forks are native */
		hpb_temporary.ioParam.ioPermssn = permissions;
		hpb_temporary.fileParam.ioFVersNum = 0;
		hpb_temporary.fileParam.ioVRefNum = srcFile->volume;
		hpb_temporary.fileParam.ioDirID = srcFile->directory;

		if (srcFile->name!=(char*)NULL)
		{
			name_temporary[0] = srcFile->length;
			BlockMoveData(srcFile->name,&name_temporary[1],srcFile->length);
		} else
			name_temporary[0] = 0;
		hpb_temporary.fileParam.ioNamePtr = name_temporary;

		if ((fork_name=srcFile->fork)==(char*)NULL || (*fork_name)==0)
		{
			if ((err=PBHOpenSync(&hpb_temporary))!=noErr)
				return translate_platform_error(err,_platform_MacOS);
		} else if (strcmp(fork_name,cs_resource_fork_name))
		{
			hpb_temporary.fileParam.ioDirID = srcFile->volume;
			if ((err=PBHOpenSync(&hpb_temporary))!=noErr)
				return translate_platform_error(err,_platform_MacOS);
		} else
		{
			if ((err=PBHOpenRFSync(&hpb_temporary))!=noErr)
				return translate_platform_error(err,_platform_MacOS);
		}

		return hpb_temporary.ioParam.ioRefNum;
	} else
	{
		FSRef		fs_reference;
		SInt16		refNum;

		if ((err=cs_file_to_fsref(srcFile,&fs_reference,permissions))!=0)
			return err;

		if ((err=FSOpenFork(&fs_reference,unistr_temporary.length,
				unistr_temporary.unicode,permissions,&refNum))!=noErr)
			return translate_platform_error(err,_platform_MacOS);

		return refNum;
	}
}

int fs_close_fork(int ref)
{
	if (file_system_info==0) evaluate_file_system();
	/* i don't see a difference, really but i'll stick it where it came out */
	if (file_system_info<=0)
	{
		pb_temporary.ioParam.ioRefNum = ref;
		return translate_platform_error(PBCloseSync(&pb_temporary),_platform_MacOS);
	} else
	{
		fsf_temporary.forkRefNum = ref;
		return translate_platform_error(PBCloseForkSync(&fsf_temporary),_platform_MacOS);
	}
}

static int cs_file_to_fsref(const cs_filespec *src, FSRef *dst, UInt8 permissions)
{
	OSErr			err;

	/* i'm not sure if this is completely necessary, but just as a precaution */
/*memset((void*)&fsf_temporary,0,sizeof(fsf_temporary));*/

	fsf_temporary.permissions = permissions;
/*fsf_temporary.ref = dst;*/

	fsr_temporary.ioVRefNum = src->volume;
	fsr_temporary.ioDirID = src->directory;
	fsr_temporary.newRef = dst;

	if (src->encoding==_text_encoding_MacRoman)
	{
		if (src->name!=(char*)NULL)
		{
			name_temporary[0] = src->length;
			BlockMoveData(src->name,&name_temporary[1],src->length);
		} else
			name_temporary[0] = 0;
		fsr_temporary.ioNamePtr = name_temporary;
		err = PBMakeFSRefSync(&fsr_temporary);
	} else if (src->encoding==_text_encoding_MacUnicode)
	{
		fsr_temporary.ioNamePtr = (StringPtr)NULL;
		if ((err=PBMakeFSRefSync(&fsr_temporary))==noErr)
			err = FSMakeFSRefUnicode(dst,src->length,src->name,kTextEncodingMacUnicode,dst);
	} else
		return kTS_encoding_not_found;

	/* forks - if empty or void, leave empty (data fork), else set to the OS's
	  RF string if equal to cs_resource_fork_name */
	if (strcmp(cs_resource_fork_name,src->fork))
	{
		err = FSGetResourceForkName(&unistr_temporary);
	} else
		return cstring_to_unistring(src->fork,&unistr_temporary);

	return translate_platform_error(err,_platform_MacOS);;
}

static OSErr copy_fork_to_non_native_file(HParmBlkPtr hpb,
	const char *fork,
	SInt64 eof)
{
	CInfoPBRec		catInfo;
	HParamBlockRec	temp_hpb;
	Str255			astr;

	Ptr				copy_buffer = (Ptr)NULL;
	SInt64			position;
	SInt32			length, file_copy_buffer_size = DEFAULT_FILE_COPY_BUFFER_SIZE;
	OSErr			err = noErr;

	cpstrcpy(astr,fork,27);
	*(SInt32*)&astr[astr[0]] = '.FRK';
	astr[0] += 4;

	catInfo.hFileInfo.ioVRefNum = hpb->fileParam.ioVRefNum;
	catInfo.hFileInfo.ioDirID = hpb->fileParam.ioDirID;
	catInfo.hFileInfo.ioNamePtr = astr;
	if ((err=PBGetCatInfoSync(&catInfo))==noErr)
		return err;

	temp_hpb.fileParam.ioVRefNum = catInfo.hFileInfo.ioVRefNum;
	temp_hpb.fileParam.ioDirID = catInfo.hFileInfo.ioDirID;
	/* we open the file in the fork allocation directory with the same name */
	temp_hpb.fileParam.ioNamePtr = hpb->fileParam.ioNamePtr;

	if ((err=PBHOpenSync(&temp_hpb))!=noErr)
		return err;
	/*if ((err=PBGetFPosSync(ParmBlkPtr)hpb))!=noErr)
		return err;*/

	SetInt64(position,0,0);

	/* reduce the buffer if memory is tight */
	while((copy_buffer=NewPtr(file_copy_buffer_size))==(Ptr)NULL)
		if ((file_copy_buffer_size>>=1)<1024)
			return kCS_memory_full_error; /* heap is rediculously full */

	hpb->ioParam.ioPosMode = temp_hpb.ioParam.ioPosMode = fsFromStart;

	while(!Int64Z(eof))
	{
		if (S64GtU32(position,SINT32_MAX))
		{
			hpb->ioParam.ioPosMode |= kUseWidePositioning;
			temp_hpb.ioParam.ioPosMode |= kUseWidePositioning;
			*(SInt64*)&((*(XIOParam*)&hpb).ioWPosOffset) = position;
			*(SInt64*)&((*(XIOParam*)&temp_hpb).ioWPosOffset) = position;
		} else
		{
			hpb->ioParam.ioPosMode &= ~kUseWidePositioning;
			temp_hpb.ioParam.ioPosMode &= ~kUseWidePositioning;
			temp_hpb.ioParam.ioPosOffset = hpb->ioParam.ioPosOffset =
				Int64Lo(position);
		}

		length = file_copy_buffer_size;
		if (U64LtU32(eof,length)) length = Int64Lo(eof);

		hpb->ioParam.ioBuffer = copy_buffer;
		hpb->ioParam.ioReqCount = length;
		if ((err=PBReadSync((ParmBlkPtr)hpb))!=noErr)
			return err;

		hpb->ioParam.ioBuffer = copy_buffer;

		assert(length==temp_hpb.ioParam.ioReqCount);

		length = temp_hpb.ioParam.ioReqCount = hpb->ioParam.ioActCount;
		if ((err=PBWriteSync((ParmBlkPtr)&temp_hpb))!=noErr)
			return err;

		Add64U32(position,length);
		Sub64U32(eof,length);
	}
	return err;
}

static int allocate_non_native_directory_fork(char *name)
{
	UInt8			*old_name;
	long			old_dir;
	int				result;

/**(UInt64*)&name_temporary[1]=*(UInt64*)name;*/
	cpstrcpy(name_temporary,name,27);
	*(SInt32*)&name_temporary[name_temporary[0]] = '.FRK';
	name_temporary[0] += 4;
	old_name = hpb_temporary.fileParam.ioNamePtr;
	old_dir = hpb_temporary.fileParam.ioDirID;
	hpb_temporary.fileParam.ioNamePtr = name_temporary;

	if ((result=PBDirCreateSync(&hpb_temporary))==noErr)
		result = hpb_temporary.fileParam.ioDirID;
	else if (result==-48) /* dupFNErr */
	{
		CInfoPBRec		catInfo;

		catInfo.hFileInfo.ioVRefNum = hpb_temporary.fileParam.ioVRefNum;
		catInfo.hFileInfo.ioDirID = old_dir;
		catInfo.hFileInfo.ioNamePtr = name_temporary;
		if ((result=PBGetCatInfoSync(&catInfo))==noErr)
			result = catInfo.hFileInfo.ioDirID;
	}

	hpb_temporary.fileParam.ioNamePtr = old_name;
	hpb_temporary.fileParam.ioDirID = old_dir;

	return result;
}

/* only needs to be called once, if necessary */
static void evaluate_file_system()
{
	SInt32			result;
	OSErr			err;

	/* worst case senario is that it doesn't support HFS+, let alone carry the
	  extended file system APIs (before MacOS 7.6.1) */
	file_system_info = -2;

	/* this should NOT return an error */
	if ((err=Gestalt(gestaltFSAttr,&result))!=noErr)
		return;

	if (result&(1<<gestaltHasHFSPlusAPIs))
		file_system_info = 1; /* everything is good */
	else if ((result&HFSPLUS_COMPATIBLE_BITS)==HFSPLUS_COMPATIBLE_BITS)
		file_system_info = -1; /* not great but good enough */
}

platform_handle fs_read_to_handle(int refNum, SInt64 offset, SInt32 length)
{
	platform_handle		data = (platform_handle)NULL;

	if (refNum!=NONE)
	{
		if ((data=new_platform_handle(length,1))!=(platform_handle)NULL)
		{
			XIOParam		file;

			file.ioCompletion = (IOCompletionUPP)NULL;
			file.ioRefNum = refNum;
			file.ioReqCount = length;
			file.ioPosMode = Int64Hi(offset)<0 ? fsFromLEOF : fsFromStart;

			if ((Int64Hi(offset)==0 && Int64Lo(offset)<=SINT32_MAX)
					|| (Int64Hi(offset)==-1 && Int64Lo(offset)>=SINT32_MIN))
			{
				/*SetInt64(file.ioWPosOffset,Int64Lo(offset),0);*/
				file.ioWPosOffset.hi = Int64Lo(offset);
			} else
			{	/* try for wide positioning anyway; if we fail, we fail */
				file.ioPosMode |= kUseWidePositioning;
				*(SInt64*)&file.ioWPosOffset = offset;
			}

			file.ioBuffer = *data;

			if (PBReadSync((ParamBlockRec*)&file)==noErr)
			{
				unlock_platform_handle(data);
			} else
			{
				dispose_platform_handle(data);
				data = (platform_handle)NULL;
			}
		}
	}
	return data;
}

static int cstring_to_unistring(char *str, HFSUniStr255 *ustr)
{
	UniChar			*d = ustr->unicode;
	char			*s;
	int				x;

	if ((s=str)==(char*)NULL || ustr==(HFSUniStr255*)NULL)
		return kCS_parameter_error;
	if ((x=ustr->length=strlen(str))>0)
		do
		{
			signed char		i;

			if ((i=*s++)<0)
				return -1; /* not implemented */
			*d++ = i;
		} while(--x);

	return 0;
}
